/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */


using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;
using BEPUphysics;
using BEPUphysics.Entities;
using BEPUphysics.ResourceManagement;

namespace BEPUphysicsDemos
{

    /// <summary>
    /// Simple camera class for moving around the demos area.
    /// </summary>
    public class Camera
    {
        /// <summary>
        /// Gets or sets the position of the camera.
        /// </summary>
        public Vector3 position { get; set; }
        float myYaw;
        float myPitch;
        /// <summary>
        /// Gets or sets the yaw rotation of the camera.
        /// </summary>
        public float yaw
        {
            get
            {
                return myYaw;
            }
            set
            {
                myYaw = MathHelper.WrapAngle(value);
            }
        }
        /// <summary>
        /// Gets or sets the pitch rotation of the camera.
        /// </summary>
        public float pitch
        {
            get
            {
                return myPitch;
            }
            set
            {
                myPitch = value;
                if (myPitch > MathHelper.PiOver2)
                    myPitch = MathHelper.PiOver2;
                else if (myPitch < -MathHelper.PiOver2)
                    myPitch = -MathHelper.PiOver2;
            }
        }

        /// <summary>
        /// Gets or sets the speed at which the camera moves.
        /// </summary>
        public float speed { get; set; }

        Matrix myViewMatrix;
        /// <summary>
        /// Gets the view matrix of the camera.
        /// </summary>
        public Matrix viewMatrix
        {
            get
            {
                return myViewMatrix;
            }
        }
        /// <summary>
        /// Gets or sets the projection matrix of the camera.
        /// </summary>
        public Matrix projectionMatrix { get; set; }

        Matrix myWorldMatrix;
        /// <summary>
        /// Gets the world transformation of the camera.
        /// </summary>
        public Matrix worldMatrix
        {
            get
            {
                return myWorldMatrix;
            }
        }
        /// <summary>
        /// Whether or not to use the default free-flight camera controls.
        /// Set to false when using vehicles or character controllers.
        /// </summary>
        public bool useMovementControls = true;


        #region Chase Camera Mode
        //The following are used for the chase camera only.
        /// <summary>
        /// Entity to follow around and point at.
        /// </summary>
        Entity entityToChase;
        /// <summary>
        /// Offset vector from the center of the target chase entity to look at.
        /// </summary>
        Vector3 offsetFromChaseTarget;
        /// <summary>
        /// Whether or not to transform the offset vector with the rotation of the entity.
        /// </summary>
        bool transformOffset;
        /// <summary>
        /// Distance away from the target entity to try to maintain.  The distance will be shorter at times if the ray hits an object.
        /// </summary>
        float distanceToTarget;
        /// <summary>
        /// Whether or not the camera is currently in chase camera mode.
        /// </summary>
        bool isChaseCameraMode = false;
        /// <summary>
        /// Sets up all the information required by the chase camera.
        /// </summary>
        /// <param name="target">Target to follow.</param>
        /// <param name="offset">Offset from the center of the entity target to point at.</param>
        /// <param name="transform">Whether or not to transform the offset with the target entity's rotation.</param>
        /// <param name="distance">Distance from the target position to try to maintain.</param>
        public void activateChaseCameraMode(Entity target, Vector3 offset, bool transform, float distance)
        {
            entityToChase = target;
            offsetFromChaseTarget = offset;
            transformOffset = transform;
            distanceToTarget = distance;
            isChaseCameraMode = true;
        }
        /// <summary>
        /// Disable the chase camera mode, returning it to first person perspective.
        /// </summary>
        public void deactivateChaseCameraMode()
        {
            isChaseCameraMode = false;
        }
        #endregion

        /// <summary>
        /// Creates a camera.
        /// </summary>
        /// <param name="pos">Initial position of the camera.</param>
        /// <param name="s">Speed of the camera per second.</param>
        /// <param name="p">Initial pitch angle of the camera.</param>
        /// <param name="y">Initial yaw value of the camera.</param>
        /// <param name="projMatrix">Projection matrix used.</param>
        public Camera(Vector3 pos, float s, float p, float y, Matrix projMatrix)
        {
            position = pos;
            speed = s;
            yaw = y;
            pitch = p;
            projectionMatrix = projMatrix;
        }

        /// <summary>
        /// Moves the camera forward.
        /// </summary>
        /// <param name="distance">Distance to move.</param>
        public void moveForward(float distance)
        {
            position += myWorldMatrix.Forward * distance;
        }

        /// <summary>
        /// Moves the camera to the right.
        /// </summary>
        /// <param name="distance">Distance to move.</param>
        public void moveRight(float distance)
        {
            position += myWorldMatrix.Right * distance;
        }

        /// <summary>
        /// Moves the camera up.
        /// </summary>
        /// <param name="distance">Distanec to move.</param>
        public void moveUp(float distance)
        {
            position += new Vector3(0, distance, 0);
        }


        /// <summary>
        /// Updates the state of the camera.
        /// </summary>
        /// <param name="dt">Time since the last frame in seconds.</param>
        /// <param name="keyboardInput">Input for this frame from the keyboard.</param>
        /// <param name="oldMouseInput">Input from the previous frame from the mouse.</param>
        /// <param name="mouseInput">Input for this frame from the mouse.</param>
        /// <param name="gamePadInput">Input for this frame from the game pad.</param>
#if !WINDOWS
        public void update(float dt, KeyboardState keyboardInput, GamePadState gamePadInput)
        {
            yaw += gamePadInput.ThumbSticks.Right.X * -1.5f * dt;
            pitch += gamePadInput.ThumbSticks.Right.Y * 1.5f * dt;

#else
        public void update(float dt, KeyboardState keyboardInput, MouseState mouseInput, GamePadState gamePadInput)
        {
            yaw += (200 - mouseInput.X) * dt * .12f;
            pitch += (200 - mouseInput.Y) * dt * .12f;
#endif

            myWorldMatrix = Matrix.CreateFromAxisAngle(Vector3.Right, pitch) * Matrix.CreateFromAxisAngle(Vector3.Up, yaw);

            if (isChaseCameraMode)
            {
                Vector3 offset;
                if (transformOffset)
                    offset = Vector3.Transform(offsetFromChaseTarget, entityToChase.orientationMatrix);
                else
                    offset = offsetFromChaseTarget;
                Vector3 lookAt = entityToChase.centerPosition + offset;
                Vector3 backwards = worldMatrix.Backward;
                List<Vector3> hitLocations = Resources.getVectorList(), hitNormals = Resources.getVectorList();
                List<float> hitTimes = Resources.getFloatList();
                List<Entity> hitEntities = Resources.getEntityList();
                //Find the earliest ray hit that isn't the chase target to position the camera appropriately.
                if (entityToChase.space.rayCast(lookAt, backwards, distanceToTarget, true, hitEntities, hitLocations, hitNormals, hitTimes))
                {
                    float earliestTime = float.PositiveInfinity;
                    int earliestHitIndex;
                    for (int k = 0; k < hitTimes.Count; k++)
                    {
                        if (hitEntities[k] == entityToChase || (hitEntities[k] != null && hitEntities[k].collisionRules.personal > CollisionRule.normal))
                        {
                            hitTimes[k] = float.PositiveInfinity;
                        }
                        else if (hitTimes[k] < earliestTime)
                        {

                            earliestTime = hitTimes[k];
                            earliestHitIndex = k;

                        }
                    }
                    if (earliestTime == float.PositiveInfinity)
                        earliestTime = distanceToTarget;
                    position = lookAt + (earliestTime) * backwards; //Put the camera just before any hit spot.
                }
                else
                    position = lookAt + (distanceToTarget) * backwards;
                Resources.giveBack(hitEntities);
                Resources.giveBack(hitLocations);
                Resources.giveBack(hitNormals);
                Resources.giveBack(hitTimes);
            }
            else if (useMovementControls)
            {//Only move around if the camera has control over its own position.
                float distance = speed * dt;
#if !WINDOWS
                moveForward(gamePadInput.ThumbSticks.Left.Y * distance);
                moveRight(gamePadInput.ThumbSticks.Left.X * distance);
                if (gamePadInput.IsButtonDown(Buttons.LeftStick))
                    moveUp(distance);
                if (gamePadInput.IsButtonDown(Buttons.RightStick))
                    moveUp(-distance);
#endif

                if (keyboardInput.IsKeyDown(Keys.E))
                    moveForward(distance);
                if (keyboardInput.IsKeyDown(Keys.D))
                    moveForward(-distance);
                if (keyboardInput.IsKeyDown(Keys.S))
                    moveRight(-distance);
                if (keyboardInput.IsKeyDown(Keys.F))
                    moveRight(distance);
                if (keyboardInput.IsKeyDown(Keys.A))
                    moveUp(distance);
                if (keyboardInput.IsKeyDown(Keys.Z))
                    moveUp(-distance);
            }

            myWorldMatrix = myWorldMatrix * Matrix.CreateTranslation(position);
            myViewMatrix = Matrix.Invert(myWorldMatrix);

        }


    }
}
