﻿/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using BEPUphysics.Entities;
using Microsoft.Xna.Framework;
using BEPUphysics.Constraints;

namespace BEPUphysicsDemos.Demos
{
    /// <summary>
    /// Poseable humanoid with joint friction.
    /// </summary>
    public class ActionFigureDemo : StandardDemo
    {
        /// <summary>
        /// Gets the name of the simulation.
        /// </summary>
        public override string name
        {
            get { return "Action Figure"; }
        }

        /// <summary>
        /// Constructs a new demo.
        /// </summary>
        /// <param name="game">Game owning this demo.</param>
        public ActionFigureDemo(DemosGame game)
            : base(game)
        {
            //Make a simple, poseable action figure.  This isn't a full featured 'ragdoll' really;
            //ragdolls usually have specific joint limits and appropriate kinds of joints rather than all
            //ball socket joints.  This demo could be modified into a 'proper' ragdoll.
            Entity body = new Box(new Vector3(0, 5, 0), 1.5f, 2, 1, 10);
            space.add(body);

            Entity head = new Sphere(body.centerPosition + new Vector3(0, 2, 0), .5f, 5);
            space.add(head);

            //Connect the head to the body.
            space.add(new BallSocketJoint(body, head, head.centerPosition + new Vector3(0, -.9f, 0)));
            //Angular motors can be used to simulate friction when their goal velocity is 0.
            AngularMotor angularMotor = new AngularMotor(body, head);
            angularMotor.settings.maximumForce = 150; //The maximum force of 'friction' in this joint.
            space.add(angularMotor);

            //Make the first arm.
            Box upperLimb = new Box(body.centerPosition + new Vector3(-1.6f, .8f, 0), 1, .5f, .5f, 5);
            space.add(upperLimb);

            Box lowerLimb = new Box(upperLimb.centerPosition + new Vector3(-1.4f, 0, 0), 1, .5f, .5f, 5);
            space.add(lowerLimb);

            //Connect the body to the upper arm.
            space.add(new BallSocketJoint(body, upperLimb, upperLimb.centerPosition + new Vector3(.7f, 0, 0)));
            angularMotor = new AngularMotor(body, upperLimb);
            angularMotor.settings.maximumForce = 250;
            space.add(angularMotor);


            //Connect the upper arm to the lower arm.
            space.add(new BallSocketJoint(upperLimb, lowerLimb, upperLimb.centerPosition + new Vector3(-.7f, 0, 0)));
            angularMotor = new AngularMotor(upperLimb, lowerLimb);
            angularMotor.settings.maximumForce = 150;
            space.add(angularMotor);

            //Make the second arm.
            upperLimb = new Box(body.centerPosition + new Vector3(1.6f, .8f, 0), 1, .5f, .5f, 5);
            space.add(upperLimb);

            lowerLimb = new Box(upperLimb.centerPosition + new Vector3(1.4f, 0, 0), 1, .5f, .5f, 5);
            space.add(lowerLimb);

            //Connect the body to the upper arm.
            space.add(new BallSocketJoint(body, upperLimb, upperLimb.centerPosition + new Vector3(-.7f, 0, 0)));
            //Angular motors can be used to simulate friction when their goal velocity is 0.
            angularMotor = new AngularMotor(body, upperLimb);
            angularMotor.settings.maximumForce = 250; //The maximum force of 'friction' in this joint.
            space.add(angularMotor);


            //Connect the upper arm to the lower arm.
            space.add(new BallSocketJoint(upperLimb, lowerLimb, upperLimb.centerPosition + new Vector3(.7f, 0, 0)));
            angularMotor = new AngularMotor(upperLimb, lowerLimb);
            angularMotor.settings.maximumForce = 150;
            space.add(angularMotor);

            //Make the first leg.
            upperLimb = new Box(body.centerPosition + new Vector3(-.6f, -2.1f, 0), .5f, 1.3f, .5f, 8);
            space.add(upperLimb);

            lowerLimb = new Box(upperLimb.centerPosition + new Vector3(0, -1.7f, 0), .5f, 1.3f, .5f, 8);
            space.add(lowerLimb);

            //Connect the body to the upper arm.
            space.add(new BallSocketJoint(body, upperLimb, upperLimb.centerPosition + new Vector3(0, .9f, 0)));
            //Angular motors can be used to simulate friction when their goal velocity is 0.
            angularMotor = new AngularMotor(body, upperLimb);
            angularMotor.settings.maximumForce = 350; //The maximum force of 'friction' in this joint.
            space.add(angularMotor);


            //Connect the upper arm to the lower arm.
            space.add(new BallSocketJoint(upperLimb, lowerLimb, upperLimb.centerPosition + new Vector3(0, -.9f, 0)));
            angularMotor = new AngularMotor(upperLimb, lowerLimb);
            angularMotor.settings.maximumForce = 250;
            space.add(angularMotor);

            //Make the second leg.
            upperLimb = new Box(body.centerPosition + new Vector3(.6f, -2.1f, 0), .5f, 1.3f, .5f, 8);
            space.add(upperLimb);

            lowerLimb = new Box(upperLimb.centerPosition + new Vector3(0, -1.7f, 0), .5f, 1.3f, .5f, 8);
            space.add(lowerLimb);

            //Connect the body to the upper arm.
            space.add(new BallSocketJoint(body, upperLimb, upperLimb.centerPosition + new Vector3(0, .9f, 0)));
            //Angular motors can be used to simulate friction when their goal velocity is 0.
            angularMotor = new AngularMotor(body, upperLimb);
            angularMotor.settings.maximumForce = 350; //The maximum force of 'friction' in this joint.
            space.add(angularMotor);


            //Connect the upper arm to the lower arm.
            space.add(new BallSocketJoint(upperLimb, lowerLimb, upperLimb.centerPosition + new Vector3(0, -.9f, 0)));
            angularMotor = new AngularMotor(upperLimb, lowerLimb);
            angularMotor.settings.maximumForce = 250;
            space.add(angularMotor);

            //Add some ground.
            space.add(new Box(new Vector3(0, -3.5f, 0), 40f, 1, 40f));

            game.camera.position = new Vector3(0, 5, 25);
        }

    }
}
