/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using BEPUphysics;
using BEPUphysics.Entities;
using BEPUphysicsDemos.SampleCode;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using BEPUphysicsDrawer.Lines;
using BEPUphysics.ResourceManagement;

namespace BEPUphysicsDemos.Demos
{
    /// <summary>
    /// Superclass of the normal demos that let the user
    /// shoot spheres, grab things, create vehicles, 
    /// and walk around as characters.
    /// </summary>
    public abstract class StandardDemo : Demo
    {
        protected CharacterControllerInput character;
        protected VehicleInput vehicle;
        protected Entity kapow;
        protected Explosion kapowMaker;
        protected GrabSpring grabber;
        protected LineDisplayObjectBase grabberGraphic;
        protected float grabDistance;

        protected Texture2D whitePixel;

        public StandardDemo(DemosGame game)
            : base(game)
        {
            //Creates the player character (C).
            character = new CharacterControllerInput(space, game.camera);
            //Creates the drivable vehicle (V).
            Model wheelModel = game.Content.Load<Model>("carWheel");
            Texture2D wheelTexture = game.Content.Load<Texture2D>("wheel");
            whitePixel = game.Content.Load<Texture2D>("whitePixel");
            vehicle = new VehicleInput(new Vector3(10000, 0, 0), space, game.camera, game.modelDrawer, wheelModel, wheelTexture);

            space.simulationSettings.motionUpdate.gravity = new Vector3(0, -9.81f, 0f); //If left unset, the default value is (0,0,0).

            //Create the tossable ball.
            kapow = new Sphere(new Vector3(11000, 0, 0), .5f, 20);
            kapow.collisionMargin = .1f;
            kapowMaker = new Explosion(Vector3.Zero, 400, 15, space);
            //Create the right-click grab spring.
            grabber = new GrabSpring(30f, .8f, .6f);
            grabber.isUpdating = false;
            grabberGraphic = game.constraintDrawer.add(grabber);
            grabberGraphic.isActive = false;
            space.add(grabber);
            space.add(kapow);

            //IMPORTANT PERFORMANCE NOTE:
            //  BEPUphysics uses an iterative system to solve constraints.  You can tell it to do more or less iterations.
            //  Less iterations is faster; more iterations makes the result more accurate.
            //
            //  The amount of iterations needed for a simulation varies.  The "Wall" and "Pyramid" simulations (#1 and #16) are each fairly
            //  solver intensive, but as few as 4 iterations can be used with acceptable results.
            //  The "Jenga" (#19) simulation usually needs a few more iterations for stability; 7-9 is a good minimum.
            //
            //  The Dogbot demo (#9) shows how accuracy can smoothly increase with more iterations.
            //  With very few iterations (1-3), it has slightly jaggier movement, as if the parts used to construct it were a little cheap.
            //  As you give it a few more iterations, the motors and constraints get more and more robust.
            //  
            //  Many simulations can work perfectly fine with very few iterations, 
            //  and using a low number of iterations can substantially improve performance.
            //
            //  To change the number of iterations used, uncomment and change the following line (10 iterations is the default):

            //space.simulationSettings.collisionResponse.iterations = 10;
        }

        public override void update(float dt)
        {

            #region Kapow-Shooter Input
            //Update kapow-shooter
            if (!vehicle.isActive)
#if !WINDOWS
                if (game.gamePadInput.IsButtonDown(Buttons.RightTrigger))
#else
                if (game.mouseInput.LeftButton == ButtonState.Pressed)
#endif
                {
                    if (character.isActive)//Keep the ball out of the character's body if its being used.
                        kapow.centerPosition = game.camera.position + game.camera.worldMatrix.Forward * 3;
                    else
                        kapow.centerPosition = game.camera.position + game.camera.worldMatrix.Forward;
                    kapow.angularVelocity = Vector3.Zero;
                    kapow.linearVelocity = game.camera.worldMatrix.Forward * 30;

                }
            #endregion
            #region Grabber Input
            //Update grabber

#if !WINDOWS
            if (game.gamePadInput.IsButtonDown(Buttons.RightShoulder) && !grabber.isUpdating)
#else
            if (game.mouseInput.RightButton == ButtonState.Pressed && !grabber.isUpdating)
#endif
            {

                Vector3 startPosition;
                if (character.isActive)
                    startPosition = game.camera.position + 5 * game.camera.worldMatrix.Forward;
                else
                    startPosition = game.camera.position;
                List<Vector3> hitLocations = Resources.getVectorList(), hitNormals = Resources.getVectorList();
                List<float> hitTimes = Resources.getFloatList();
                List<Entity> hitEntities = Resources.getEntityList();

                //Find the earliest ray hit

                if (space.rayCast(startPosition, game.camera.worldMatrix.Forward, 1000, true, hitEntities, hitLocations, hitNormals, hitTimes))
                {
                    float earliestTime = float.PositiveInfinity;
                    int earliestHitIndex = 0;
                    for (int k = 0; k < hitTimes.Count; k++)
                    {

                        if (hitEntities[k] != null && hitEntities[k].collisionRules.personal > CollisionRule.normal)
                        {//Don't bother selecting intangible/detector objects.
                            hitTimes[k] = float.PositiveInfinity;
                        }
                        else if (hitTimes[k] < earliestTime)
                        {

                            earliestTime = hitTimes[k];
                            earliestHitIndex = k;

                        }
                    }
                    //If there's a valid ray hit, then grab the connected object!
                    if (hitEntities[earliestHitIndex] != null && hitTimes[earliestHitIndex] != float.PositiveInfinity && hitEntities[earliestHitIndex].isDynamic)
                    {
                        grabber.setup(hitEntities[earliestHitIndex], hitLocations[earliestHitIndex]);
                        grabber.isUpdating = true;
                        grabberGraphic.isActive = true;
                        if (character.isActive)
                            grabDistance = hitTimes[earliestHitIndex] + 5;
                        else
                            grabDistance = hitTimes[earliestHitIndex];

                    }
                }


            }
#if !WINDOWS
            if (game.gamePadInput.IsButtonDown(Buttons.RightShoulder) && grabber.isUpdating)
#else
            else if (game.mouseInput.RightButton == ButtonState.Pressed && grabber.isUpdating)
#endif
            {
                grabber.goalPosition = game.camera.position + game.camera.worldMatrix.Forward * grabDistance;
            }
#if !WINDOWS
            if (!game.gamePadInput.IsButtonDown(Buttons.RightShoulder) && grabber.isUpdating)
#else
            else if (game.mouseInput.RightButton == ButtonState.Released && grabber.isUpdating)
#endif
            {
                grabber.isUpdating = false;
                grabberGraphic.isActive = false;
            }
            #endregion
            #region Control State Input
#if !WINDOWS
            if (!vehicle.isActive && game.wasButtonPressed(Buttons.LeftTrigger))
            {
                kapowMaker.position = kapow.centerPosition;
                kapowMaker.explode();
            }
            if (game.wasButtonPressed(Buttons.A))
            {//Toggle character perspective.
                if (!character.isActive)
                {
                    vehicle.deactivate();
                    character.activate();
                }
                else
                    character.deactivate();
            }
            if (game.wasButtonPressed(Buttons.B))
            {//Toggle vehicle perspective.
                if (!vehicle.isActive)
                {
                    character.deactivate();
                    vehicle.activate();
                }
                else
                    vehicle.deactivate();
            }
#else
            if (!vehicle.isActive && game.wasKeyPressed(Keys.Space))
            {//Detonate the bomb
                kapowMaker.position = kapow.centerPosition;
                kapowMaker.explode();
            }
            if (game.wasKeyPressed(Keys.C))
            {//Toggle character perspective.
                if (!character.isActive)
                {
                    vehicle.deactivate();
                    character.activate();
                }
                else
                    character.deactivate();
            }


            if (game.wasKeyPressed(Keys.V))
            {//Toggle vehicle perspective.
                if (!vehicle.isActive)
                {
                    character.deactivate();
                    vehicle.activate();
                }
                else
                    vehicle.deactivate();
            }
#endif
            #endregion

            base.update(dt); //Base.update updates the space, which needs to be done before the character or vehicle are updated.

            character.update(space.simulationSettings.timeStep.timeStepDuration, game.previousKeyboardInput, game.keyboardInput, game.previousGamePadInput, game.gamePadInput);
            vehicle.update(space.simulationSettings.timeStep.timeStepDuration, game.keyboardInput, game.gamePadInput);

        }

        public override void cleanUp()
        {
            //Wouldn't want the character or vehicle to own the camera after we switch.
            character.deactivate();
            vehicle.deactivate();
        }

        public override void drawUI()
        {
#if !WINDOWS
            if (game.gamePadInput.IsButtonDown(Buttons.RightShoulder))
#else
            if (game.mouseInput.RightButton == ButtonState.Pressed)
#endif
                game.UIDrawer.Draw(whitePixel, new Rectangle(game.graphics.PreferredBackBufferWidth / 2, game.graphics.PreferredBackBufferHeight / 2, 3, 3), Color.LightBlue);
        }
    }
}