﻿/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using BEPUphysics.Entities;
using Microsoft.Xna.Framework;
using BEPUphysics.Constraints;
using BEPUphysics;

namespace BEPUphysicsDemos.Demos
{
    /// <summary>
    /// A basic lattice of constraints acting like cloth.
    /// </summary>
    public class ClothDemo : StandardDemo
    {
        /// <summary>
        /// Gets the name of the simulation.
        /// </summary>
        public override string name
        {
            get { return "Cloth Lattice"; }
        }

        /// <summary>
        /// Constructs a new demo.
        /// </summary>
        /// <param name="game">Game owning this demo.</param>
        public ClothDemo(DemosGame game)
            : base(game)
        {
            //Joints can also act like springs by modifying their springSettings.
            //Though using a bunch of DistanceJoint objects can be slower than just applying direct spring forces,
            //it is significantly more stable and allows rigid structures.
            //The extra stability can make it useful for cloth-like simulations.
            Entity latticePiece;
            DistanceJoint distanceJoint;

            int numColumns = 16;
            int numRows = 16;
            float xSpacing = 3.09f;
            float ySpacing = 3.08f;
            float zSpacing = 3.09f;
            Entity[,] lattice = new Entity[numRows, numColumns];
            for (int i = 0; i < numRows; i++)
                for (int j = 0; j < numColumns; j++)
                {
                    latticePiece = new Sphere(
                        new Vector3(
                            xSpacing * i - (numRows - 1) * xSpacing / 2f,
                            15.58f,
                            2 + zSpacing * j - (numColumns - 1) * zSpacing / 2f),
                        .3f, 10);

                    latticePiece.localInertiaTensorInverse = Toolbox.zeroMatrix;
                    latticePiece.tag = "noDisplayObject"; //The joint lines are visible enough; don't add a sphere model for this sphere.
                    lattice[i, j] = latticePiece;

                    space.add(latticePiece);

                }

            for (int i = 0; i < numRows; i++)
                for (int j = 0; j < numColumns; j++)
                {
                    if (i + 1 < numRows)
                    {
                        distanceJoint = new DistanceJoint(lattice[i, j], lattice[i + 1, j], lattice[i, j].centerPosition, lattice[i + 1, j].centerPosition);
                        distanceJoint.springSettings.stiffnessConstant = 1000;
                        distanceJoint.springSettings.dampingConstant = 1000;
                        space.add(distanceJoint);

                        if (j + 1 < numColumns)
                        {
                            distanceJoint = new DistanceJoint(lattice[i, j], lattice[i + 1, j + 1], lattice[i, j].centerPosition, lattice[i + 1, j + 1].centerPosition);
                            distanceJoint.springSettings.stiffnessConstant = 1000;
                            distanceJoint.springSettings.dampingConstant = 1000;
                            space.add(distanceJoint);
                        }
                        if (j - 1 >= 0)
                        {
                            distanceJoint = new DistanceJoint(lattice[i, j], lattice[i + 1, j - 1], lattice[i, j].centerPosition, lattice[i + 1, j - 1].centerPosition);
                            distanceJoint.springSettings.stiffnessConstant = 1000;
                            distanceJoint.springSettings.dampingConstant = 1000;
                            space.add(distanceJoint);
                        }
                    }
                    if (j + 1 < numColumns)
                    {
                        distanceJoint = new DistanceJoint(lattice[i, j], lattice[i, j + 1], lattice[i, j].centerPosition, lattice[i, j + 1].centerPosition);
                        distanceJoint.springSettings.stiffnessConstant = 1000;
                        distanceJoint.springSettings.dampingConstant = 1000;
                        space.add(distanceJoint);
                    }
                }

            //Larger amounts of iterations make the simulation more accurate, but it costs more.
            //For these sorts of solver-intensive simulations, you may find low values (1-3) provide 'good enough' results.
            space.simulationSettings.collisionResponse.iterations = 2;



            //Add some ground.
            space.add(new Sphere(new Vector3(0, 0, 0), 10));
            space.add(new Box(new Vector3(0, -3.5f, 0), 80f, 1, 80f));

            game.camera.position = new Vector3(0, 5, 25);

        }

    }
}
