﻿/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using BEPUphysics.Entities;
using Microsoft.Xna.Framework;

namespace BEPUphysicsDemos.Demos
{
    /// <summary>
    /// Non-standard shapes like MinkowskiSums, WrappedBodies, and ConvexHulls.
    /// </summary>
    public class FancyShapesDemo : StandardDemo
    {
        /// <summary>
        /// Gets the name of the simulation.
        /// </summary>
        public override string name
        {
            get { return "Fancy Shapes"; }
        }

        /// <summary>
        /// Constructs a new demo.
        /// </summary>
        /// <param name="game">Game owning this demo.</param>
        public FancyShapesDemo(DemosGame game)
            : base(game)
        {
            List<Vector3> points = new List<Vector3>();

            //Setup a random distribution in a cube and compute a convex hull.
            Random random = new Random();
            for (int k = 0; k < 40; k++)
            {
                points.Add(new Vector3(3 * (float)random.NextDouble(), 7 + 5 * (float)random.NextDouble(), 3 * (float)random.NextDouble()));
            }
            ConvexHull convexHull = new ConvexHull(points, 10);

            space.add(convexHull);

            points.Clear();

            //Create another random distribution, but this time with more points.
            points.Clear();
            for (int k = 0; k < 400; k++)
            {
                points.Add(new Vector3(4 + 1 * (float)random.NextDouble(), 7 + 3 * (float)random.NextDouble(), 1 * (float)random.NextDouble()));
            }
            convexHull = new ConvexHull(points, 10);
            space.add(convexHull);


            //Minkowski Sums are fancy 'combinations' of objects, where the result is the sum of the individual points making up shapes.
            //Think of it as sweeping one shape around and through another; a sphere and a box would produce a rounded-edge box.
            //More examples can be seen below.  Note in the third example how a minkowski sum is based on another minkowski sum.
            space.add(new MinkowskiSum(new Vector3(0, 0, 0), new Box(Vector3.Zero, 1, 1, 1), new Cone(Vector3.Zero, 2, 2), 100));
            space.add(new MinkowskiSum(new Vector3(0, 3, 0), new Cone(Vector3.Zero, 1f, 1), new Triangle(Vector3.Zero, Vector3.Right, Vector3.Forward), 100));
            space.add(new MinkowskiSum(new Vector3(0, 6, 0), new MinkowskiSum(new Vector3(10, 0, 0), new Cone(Vector3.Zero, 1f, 1), new Triangle(Vector3.Zero, Vector3.Up, Vector3.Forward * 2)), new Box(Vector3.Zero, 1, 1, 1), 50));
            space.add(new MinkowskiSum(new Vector3(0, 12, 0), new Cylinder(Vector3.Zero, 1, 2), new Triangle(new Vector3(1, 1, 1), new Vector3(-2, 0, 0), new Vector3(0, -1, 0)), 50));

            //Wrapped objects use an implicit convex hull around a set of shapes.

            //Oblique cone:
            List<Entity> cone = new List<Entity>();
            Cylinder coneBase = new Cylinder(new Vector3(0, 0, 0), 0, 1f);
            Sphere conePoint = new Sphere(new Vector3(1, 2, 0), 0f);
            cone.Add(coneBase);
            cone.Add(conePoint);
            space.add(new WrappedBody(new Vector3(-3, 0, 0), cone, 50));

            //Rather odd shape:
            List<Entity> oddShape = new List<Entity>();
            Sphere bottom = new Sphere(new Vector3(-2, 2, 0), 2f);
            Cylinder middle = new Cylinder(new Vector3(-2, 3, 0), 0, 3);
            middle.orientationQuaternion = Quaternion.CreateFromAxisAngle(Vector3.Right, (float)Math.PI / 6);
            Sphere top = new Sphere(new Vector3(-2, 4, 0), 1f);
            oddShape.Add(bottom);
            oddShape.Add(middle);
            oddShape.Add(top);
            space.add(new WrappedBody(new Vector3(-3, 4, 0), oddShape, 100));




            space.add(new Box(new Vector3(0, -10, 0), 70, 5, 70));

            game.camera.position = new Vector3(0, 0, 30);
        }

    }
}
