﻿/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using BEPUphysics.Entities;
using Microsoft.Xna.Framework;
using BEPUphysics;
using BEPUphysics.Paths;
using BEPUphysics.Paths.PathFollowing;

namespace BEPUphysicsDemos.Demos
{
    /// <summary>
    /// Entities following fancy paths.
    /// </summary>
    public class PathFollowingDemo : StandardDemo
    {
        /// <summary>
        /// Gets the name of the simulation.
        /// </summary>
        public override string name
        {
            get { return "Path Following"; }
        }

        /// <summary>
        /// Incremented and used to evaluate paths in the update method. 
        /// </summary>
        double pathTime;

        /// <summary>
        /// Defines target positions over time.
        /// </summary>
        Path<Vector3> positionPath;

        /// <summary>
        /// Defines target orientations over time.
        /// </summary>
        Path<Quaternion> orientationPath;

        /// <summary>
        /// Moves the entity.
        /// </summary>
        EntityMover mover;

        /// <summary>
        /// Rotates the entity.
        /// </summary>
        EntityRotator rotator;

        /// <summary>
        /// Constructs a new demo.
        /// </summary>
        /// <param name="game">Game owning this demo.</param>
        public PathFollowingDemo(DemosGame game)
            : base(game)
        {

            Entity movingEntity;
            //The moving entity can be either kinematic or dynamic; the EntityMover/Rotator works for either.
            movingEntity = new Box(new Vector3(-10, 0, -10), 3, 1, 1);

            //We're going to use a speed-controlled curve that wraps another curve.
            //This is the internal curve.
            //Speed-controlled curves let the user specify speeds at which an evaluator
            //will move along the curve.  Non-speed controlled curves can move at variable
            //rates based on the interpolation.
            CardinalSpline3D wrappedPositionCurve = new CardinalSpline3D();

            //Since the endpoints don't overlap, just reverse direction when they're hit.
            //The default is wrapping around; if there's a distance between the starting
            //and ending endpoints, the entity will jump very quickly (smashing anything in the way).
            wrappedPositionCurve.preLoop = CurveEndpointBehavior.Mirror;
            wrappedPositionCurve.postLoop = CurveEndpointBehavior.Mirror;

            //Start the curve up above the blocks.
            //There's two control points because the very first and very last control points
            //aren't actually reached by the curve in a CardinalSpline3D; they are used
            //to define the tangents on the interior points.
            wrappedPositionCurve.controlPoints.add(-1, new Vector3(0, 30, 0));
            wrappedPositionCurve.controlPoints.add(0f, new Vector3(0, 20, 0));
            //Add a bunch of random control points to the curve.
            Random random = new Random();
            for (int i = 1; i <= 10; i++)
            {
                wrappedPositionCurve.controlPoints.add(i, new Vector3(
                    (float)random.NextDouble() * 20 - 10,
                    (float)random.NextDouble() * 12,
                    (float)random.NextDouble() * 20 - 10));

            }

            positionPath = wrappedPositionCurve;

            //There's also a constant speed and variable speed curve type that can be used.
            //Try the following instead to move the entity at a constant rate:
            //positionPath = new ConstantLinearSpeedCurve(5, wrappedPositionCurve);

            QuaternionSlerpCurve slerpCurve = new QuaternionSlerpCurve();
            slerpCurve.controlPoints.add(0, Quaternion.Identity);
            slerpCurve.controlPoints.add(1, Quaternion.CreateFromAxisAngle(Vector3.Up, MathHelper.PiOver2));
            slerpCurve.controlPoints.add(2, Quaternion.CreateFromAxisAngle(Vector3.Up, MathHelper.Pi));
            slerpCurve.controlPoints.add(3, Quaternion.CreateFromAxisAngle(Vector3.Up, 3 * MathHelper.PiOver2));
            slerpCurve.controlPoints.add(4, Quaternion.Identity);

            slerpCurve.postLoop = CurveEndpointBehavior.Mirror;

            orientationPath = slerpCurve;


            mover = new EntityMover(movingEntity);
            //Offset the place that the mover tries to reach a little.
            //Now, when the entity spins, it acts more like a hammer swing than a saw.
            mover.localOffset = new Vector3(3, 0, 0);
            rotator = new EntityRotator(movingEntity);

            //Add the entity and movers to the space.
            space.add(movingEntity);
            space.add(mover);
            space.add(rotator);

            //Add some extra stuff to the space.
            space.add(new Box(new Vector3(0, -5, 0), 25, 10, 25));

            int numColumns = 7;
            int numRows = 7;
            int numHigh = 3;
            float xSpacing = 2.09f;
            float ySpacing = 2.08f;
            float zSpacing = 2.09f;
            for (int i = 0; i < numRows; i++)
                for (int j = 0; j < numColumns; j++)
                    for (int k = 0; k < numHigh; k++)
                    {
                        space.add(new Box(new Vector3(
                            xSpacing * i - (numRows - 1) * xSpacing / 2f,
                            1.58f + k * (ySpacing),
                            2 + zSpacing * j - (numColumns - 1) * zSpacing / 2f),
                            2, 2, 2, 10));

                    }

            game.camera.position = new Vector3(0, 5, 30);
        }

        public override void update(float dt)
        {
            //Increment the time.  Note that the space's timestep is used
            //instead of the method's dt.  This is because the demos, by
            //default, update the space once each game update.  Using the
            //space's update time keeps things synchronized.
            pathTime += space.simulationSettings.timeStep.timeStepDuration;
            mover.targetPosition = positionPath.evaluate(pathTime);
            rotator.targetOrientation = orientationPath.evaluate(pathTime);

            base.update(dt);
        }

    }
}
