﻿/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */
using System.Collections.Generic;
using BEPUphysics.Entities;
using Microsoft.Xna.Framework;
using BEPUphysics.Constraints;
using Microsoft.Xna.Framework.Input;
using BEPUphysics;
using BEPUphysics.Paths.PathFollowing;
using System.Diagnostics;
using BEPUphysics.MathExtensions;
using BEPUphysics.Paths;
using System;

namespace BEPUphysicsDemos.Demos
{
    /// <summary>
    /// For random tests and general fiddling.
    /// </summary>
    /// <remarks>
    /// This demo type is initially excluded from the main list in the DemosGame.
    /// To access it while playing the demos, add an entry to the demoTypes array for this TestDemo.
    /// </remarks>
    public class ReverseTrikeDemo : StandardDemo
    {
        /// <summary>
        /// Gets the name of the simulation.
        /// </summary>
        public override string name
        {
            get { return "Reverse Trike"; }
        }

        RevoluteMotor drivingMotor1, drivingMotor2, steeringMotor1, steeringMotor2;
        float maximumTurnAngle = MathHelper.Pi * .3f;
        float driveSpeed = 30;

        /// <summary>
        /// Constructs a new demo.
        /// </summary>
        /// <param name="game">Game owning this demo.</param>
        public ReverseTrikeDemo(DemosGame game)
            : base(game)
        {
            game.camera.position = new Vector3(0, 2, 15);
            game.camera.yaw = 0;
            game.camera.pitch = 0;

            space.add(new Box(new Vector3(0, -5, 0), 20, 1, 20));

            var body = new Box(new Vector3(0, 0, 0), 2, 1, 3, 10);
            body.centerOfMassOffset = new Vector3(0, -.8f, 0);
            space.add(body);

            #region First Wheel
            var wheel = new Cylinder(body.centerPosition + new Vector3(-1.3f, -.8f, -1.5f), .2f, .5f, 4);
            wheel.dynamicFriction = 1.5f;
            wheel.staticFriction = 1.5f;
            wheel.orientationQuaternion = Quaternion.CreateFromAxisAngle(Vector3.Forward, MathHelper.PiOver2);

            //Preventing the occasional pointless collision pair can speed things up.
            body.collisionRules.specificEntities.Add(wheel, CollisionRule.noPair);

            //Connect the wheel to the body.
            var ballSocketJoint = new BallSocketJoint(body, wheel, wheel.centerPosition);
            var swivelHingeAngularJoint = new SwivelHingeAngularJoint(body, wheel, Vector3.Up, Vector3.Right);
            //Motorize the wheel.
            drivingMotor1 = new RevoluteMotor(body, wheel, Vector3.Left);
            drivingMotor1.settings.velocityMotor.softness = .2f;
            //Let it roll when the user isn't giving specific commands.
            drivingMotor1.isActive = false;
            steeringMotor1 = new RevoluteMotor(body, wheel, Vector3.Up);
            steeringMotor1.settings.mode = MotorMode.servomechanism;
            //The constructor makes a guess about how to set up the constraint.
            //It can't always be right since it doesn't have all the information;
            //in this case, it chooses the basis and test axis incorrectly.
            //This leads to a 'flipping' behavior when the wheel is rolling
            //(the test axis is 'rolling' with the wheel, and passes over
            //a singularity which causes a flip).

            //To fix this, we configure the constraint directly.
            //The basis is aligned with how the wheel is set up; we choose 'up' as 
            //the motorized axis, and right/forward to define the angle measurement plane.
            //The test axis is set to be perpendicular to the wheel's rotation so that
            //it only measures the steering angle.

            //If you're curious, the angle measurement is just a Math.Atan2.
            //The current world test axis is dotted against the two plane axes (Right and Forward here).
            //This gives an x and y value.  These can be plugged into Atan2 just like when
            //you compute an angle on a normal 2d graph.
            steeringMotor1.basis.setWorldAxes(Vector3.Up, Vector3.Right, Vector3.Forward);
            steeringMotor1.testAxis = Vector3.Right;


            //Add the wheel and connection to the space.
            space.add(wheel);
            space.add(ballSocketJoint);
            space.add(swivelHingeAngularJoint);
            space.add(drivingMotor1);
            space.add(steeringMotor1);
            #endregion

            #region Second Wheel
            wheel = new Cylinder(body.centerPosition + new Vector3(1.3f, -.8f, -1.5f), .2f, .5f, 4);
            wheel.dynamicFriction = 1.5f;
            wheel.staticFriction = 1.5f;
            wheel.orientationQuaternion = Quaternion.CreateFromAxisAngle(Vector3.Forward, MathHelper.PiOver2);


            //Preventing the occasional pointless collision pair can speed things up.
            body.collisionRules.specificEntities.Add(wheel, CollisionRule.noPair);

            //Connect the wheel to the body.
            ballSocketJoint = new BallSocketJoint(body, wheel, wheel.centerPosition);
            swivelHingeAngularJoint = new SwivelHingeAngularJoint(body, wheel, Vector3.Up, Vector3.Right);
            //Motorize the wheel.
            drivingMotor2 = new RevoluteMotor(body, wheel, Vector3.Left);
            drivingMotor2.settings.velocityMotor.softness = .2f;
            //Let it roll when the user isn't giving specific commands.
            drivingMotor2.isActive = false;
            steeringMotor2 = new RevoluteMotor(body, wheel, Vector3.Up);
            steeringMotor2.settings.mode = MotorMode.servomechanism;
            //Configure the motor.  See wheel 1 for more description.
            steeringMotor2.basis.setWorldAxes(Vector3.Up, Vector3.Right, Vector3.Forward);
            steeringMotor2.testAxis = Vector3.Right;


            //Add the wheel and connection to the space.
            space.add(wheel);
            space.add(ballSocketJoint);
            space.add(swivelHingeAngularJoint);
            space.add(drivingMotor2);
            space.add(steeringMotor2);
            #endregion

            #region Third Wheel
            wheel = new Cylinder(body.centerPosition + new Vector3(0, -1.1f, 1.5f), .2f, .5f, 4);
            wheel.dynamicFriction = 1.5f;
            wheel.staticFriction = 1.5f;
            wheel.orientationQuaternion = Quaternion.CreateFromAxisAngle(Vector3.Forward, MathHelper.PiOver2);

            //Preventing the occasional pointless collision pair can speed things up.
            body.collisionRules.specificEntities.Add(wheel, CollisionRule.noPair);

            //Connect the wheel to the body.
            ballSocketJoint = new BallSocketJoint(body, wheel, wheel.centerPosition);
            //Notice that the third wheel isn't a swivel hinge, it's just a revolute axis.
            //This lets it roll, but prevents flopping around like the wheels of a grocery cart.
            //Could have used a RevoluteJoint solver group here, but this shows it's possible to do
            //the same things without using the combo-constraints.
            var revoluteAngularJoint = new RevoluteAngularJoint(body, wheel, Vector3.Right);

            //Add the wheel and connection to the space.
            space.add(wheel);
            space.add(ballSocketJoint);
            space.add(revoluteAngularJoint);
            #endregion

            int xLength = 256;
            int yLength = 256;

            float xSpacing = 8f;
            float ySpacing = 8f;
            float[,] heights = new float[xLength, yLength];
            float x, y;
            for (int i = 0; i < xLength; i++)
            {
                for (int j = 0; j < yLength; j++)
                {
                    x = i - xLength / 2;
                    y = j - yLength / 2;
                    //heights[i, j] = (float)(x * y / 1000f);
                    heights[i, j] = (float)(10 * (Math.Sin(x / 8) + Math.Sin(y / 8)));
                    //heights[i, j] = 3 * (float)Math.Sin(x * y / 100f);
                    //heights[i, j] = (x * x * x * y - y * y * y * x) / 1000f;

                }
            }
            //Create the terrain.
            Terrain terrain = new Terrain(new Vector3(-xLength * xSpacing / 2, -10, -yLength * ySpacing / 2));
            terrain.useFaceNormalWithinAngle = (float)Math.PI / 2; 
            terrain.setData(heights, QuadFormats.lowerLeftUpperRight, xSpacing, ySpacing);
            space.add(terrain);

            game.modelDrawer.add(terrain);


        }

        public override void update(float dt)
        {
            if (game.keyboardInput.IsKeyDown(Keys.P))
                Debug.WriteLine("Break.");

            if (game.keyboardInput.IsKeyDown(Keys.NumPad8))
            {
                //Go forward
                drivingMotor1.settings.velocityMotor.goalVelocity = driveSpeed;
                drivingMotor2.settings.velocityMotor.goalVelocity = driveSpeed;
                //The driving motors are disabled when no button is pressed, so need to turn it on.
                drivingMotor1.isActive = true;
                drivingMotor2.isActive = true;
            }
            else if (game.keyboardInput.IsKeyDown(Keys.NumPad5))
            {
                //Go backward
                drivingMotor1.settings.velocityMotor.goalVelocity = -driveSpeed;
                drivingMotor2.settings.velocityMotor.goalVelocity = -driveSpeed;
                //The driving motors are disabled when no button is pressed, so need to turn it on.
                drivingMotor1.isActive = true;
                drivingMotor2.isActive = true;
            }
            else
            {
                //Let it roll.
                drivingMotor1.isActive = false;
                drivingMotor2.isActive = false;
            }

            if (game.keyboardInput.IsKeyDown(Keys.NumPad4))
            {
                //Turn left
                steeringMotor1.settings.servo.goal = maximumTurnAngle;
                steeringMotor2.settings.servo.goal = maximumTurnAngle;
            }
            else if (game.keyboardInput.IsKeyDown(Keys.NumPad6))
            {
                //Turn right
                steeringMotor1.settings.servo.goal = -maximumTurnAngle;
                steeringMotor2.settings.servo.goal = -maximumTurnAngle;
            }
            else
            {
                //Face forward
                steeringMotor1.settings.servo.goal = 0;
                steeringMotor2.settings.servo.goal = 0;

            }

            base.update(dt);

        }

        public override void drawUI()
        {
            base.drawUI();

        }



    }
}
