﻿/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using BEPUphysics.Entities;
using Microsoft.Xna.Framework;
using BEPUphysics.Constraints;

namespace BEPUphysicsDemos.Demos
{
    /// <summary>
    /// A spinning blade connected to a piston connected to a spinning box slices up some other boxes.
    /// </summary>
    public class SawContraptionDemo : StandardDemo
    {
        /// <summary>
        /// Gets the name of the simulation.
        /// </summary>
        public override string name
        {
            get { return "Saw Contraption"; }
        }

        /// <summary>
        /// Constructs a new demo.
        /// </summary>
        /// <param name="game">Game owning this demo.</param>
        public SawContraptionDemo(DemosGame game)
            : base(game)
        {
            //Add a kinematic entity that will form the base for the strange contraption.
            Box pistonGroundAttachment = new Box(new Vector3(0, -1, 0), 1, 2, .5f);
            pistonGroundAttachment.angularVelocity = new Vector3(0, .2f, 0); //Make it spin a little to rotate the whole thing.
            space.add(pistonGroundAttachment);

            Box pistonBox1 = new Box(pistonGroundAttachment.centerPosition + new Vector3(0f, 0, 1), 1, 1, 1, 100);
            Box pistonBox2 = new Box(pistonBox1.centerPosition + new Vector3(0, 2, 0), .5f, .8f, .5f, 10);
            space.add(pistonBox1);
            space.add(pistonBox2);

            //Connect the piston entities to the base with a revolute joint that acts like an axis joint
            RevoluteJoint axisJoint = new RevoluteJoint(pistonGroundAttachment, pistonBox1, (pistonGroundAttachment.centerPosition + pistonBox1.centerPosition) / 2, Vector3.Forward);

            //Keep the axis from rotating too far so that the saw blade won't just continually ram into the ground. 
            //The limit's 'basis' and test axis used to determine the position of the limits and the current angle of the constraint
            //are automatically initialized by the RevoluteJoint constructor.  The basis and testAxis can be set afterwards; 
            //the initialization is just a reasonable 'guess' at the kind of limit configuration desired.
            axisJoint.limit.isActive = true;
            axisJoint.limit.minimumAngle = -MathHelper.PiOver2;
            axisJoint.limit.maximumAngle = MathHelper.PiOver2;
            space.add(axisJoint);

            PrismaticJoint piston = new PrismaticJoint(pistonBox1, pistonBox2, pistonBox1.centerPosition, Vector3.Up, pistonBox2.centerPosition);
            //Set up the piston limits.
            piston.limit.isActive = true; //By default, the limit and motor are both inactive.
            piston.limit.minimum = 2;
            piston.limit.maximum = 5;

            //Set up the servo motor.
            piston.motor.isActive = true;
            piston.motor.settings.mode = MotorMode.servomechanism;
            //Distance from the anchor that the piston will try to reach.
            piston.motor.settings.servo.goal = 5;
            //Set the maximum force the motor can use to reach its goal.
            piston.motor.settings.maximumForce = 100;
            //This piston, by default, moves at a constant speed, but...
            piston.motor.settings.servo.baseCorrectiveSpeed = 1;
            //... if the stiffness constant is changed to a positive value, it can also act like a spring.
            piston.motor.settings.servo.springSettings.stiffnessConstant = 0;
            //For a non-springy constraint like the piston, the dampingConstant can also be thought of as inverse 'softness.'
            piston.motor.settings.servo.springSettings.dampingConstant = 1000;

            //Add the piston to the space.
            space.add(piston);

            //Create a saw bladey object on the end of the piston.
            Box blade = new Box(pistonBox2.centerPosition + new Vector3(0, 0, .5f), .3f, 2.5f, .1f, 5);
            space.add(blade);

            //Connect the saw to the piston with a second axis joint.
            axisJoint = new RevoluteJoint(pistonBox2, blade, (pistonBox2.centerPosition + blade.centerPosition) / 2, Vector3.Forward);
            //Revolute joints can be used to make axis joints (as it is here), but you can also use them to make hinges.
            space.add(axisJoint);

            //Make the blade spin.
            axisJoint.motor.isActive = true;
            axisJoint.motor.settings.velocityMotor.goalVelocity = 30;
            axisJoint.motor.settings.maximumForce = 200;

            //Add some ground.
            space.add(new Box(new Vector3(0, -3f, 0), 20f, 1, 20f));

            //Make some debris for the saw to chop.
            for (double k = 0; k < Math.PI * 2; k += Math.PI / 20)
            {
                space.add(new Box(new Vector3((float)Math.Cos(k) * 4f, 1, (float)Math.Sin(k) * 6.5f), .5f, 1f, .5f, 10));
            }


            game.camera.position = new Vector3(0, 2, 20);
        }

    }
}
