﻿/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using BEPUphysics.Entities;
using Microsoft.Xna.Framework;
using BEPUphysics.Constraints;
using BEPUphysics;

namespace BEPUphysicsDemos.Demos
{
    /// <summary>
    /// A one-armed guy flops on the ground.  Involves a detailed ragdoll-like limb with appropriate limits.
    /// </summary>
    public class UnfortunateGuyDemo : StandardDemo
    {
        /// <summary>
        /// Gets the name of the simulation.
        /// </summary>
        public override string name
        {
            get { return "Unfortunate One-armed Guy"; }
        }

        /// <summary>
        /// Constructs a new demo.
        /// </summary>
        /// <param name="game">Game owning this demo.</param>
        public UnfortunateGuyDemo(DemosGame game)
            : base(game)
        {
            Entity ground = new Box(Vector3.Zero, 10, 1, 10);
            space.add(ground);

            //Rather than add basically redundant code for every limb, this demo
            //focuses on a single arm, showing some extra details and limits.

            //Make the torso.
            CompoundBody torso = new CompoundBody();
            torso.addBody(new Box(new Vector3(-1, 3, 0), 2, 1.5f, 1, 50));
            torso.addBody(new Sphere(new Vector3(.4f, 3, 0), .45f, 1));
            torso.addBody(new Sphere(new Vector3(-1.9f, 3.5f, 0), .25f, 1));
            torso.addBody(new Sphere(new Vector3(-1.9f, 2.5f, 0), .25f, 1));
            torso.addBody(new Sphere(new Vector3(-.3f, 2.3f, 0), .25f, 1));
            space.add(torso);

            //Make the upper arm.
            Entity upperArm = new Box(torso.centerPosition + new Vector3(1, 1.4f, 0), .4f, 1.2f, .4f, 8);
            space.add(upperArm);



            //A ball socket joint will handle the linear degrees of freedom between the two entities.
            BallSocketJoint ballSocketJoint = new BallSocketJoint(torso, upperArm, torso.centerPosition + new Vector3(1, .7f, 0));
            space.add(ballSocketJoint);

            //Shoulders don't have a simple limit.  The EllipseSwingLimit allows angles within an ellipse, which is closer to how some joints function
            //than just flat planes (like two RevoluteLimits) or a single angle (like SwingLimits).
            EllipseSwingLimit swingLimit = new EllipseSwingLimit(torso, upperArm, Vector3.Up, MathHelper.PiOver2, MathHelper.PiOver4 * 3);
            space.add(swingLimit);

            //Upper arms can't spin around forever.
            TwistLimit twistLimit = new TwistLimit(torso, upperArm, Vector3.Up, Vector3.Up, -MathHelper.PiOver4 / 2, MathHelper.PiOver4);
            twistLimit.springSettings.stiffnessConstant = 100;
            twistLimit.springSettings.dampingConstant = 100;
            space.add(twistLimit);


            //Make the lower arm.
            Entity lowerArm = new Box(upperArm.centerPosition + new Vector3(0, 1.4f, 0), .35f, 1.3f, .35f, 8);
            space.add(lowerArm);


            SwivelHingeJoint elbow = new SwivelHingeJoint(upperArm, lowerArm, upperArm.centerPosition + new Vector3(0, .6f, 0), Vector3.Forward);
            //Forearm can twist a little.
            elbow.twistLimit.isActive = true;
            elbow.twistLimit.minimumAngle = -MathHelper.PiOver4 / 2;
            elbow.twistLimit.maximumAngle = MathHelper.PiOver4 / 2;
            elbow.twistLimit.springSettings.dampingConstant = 100;
            elbow.twistLimit.springSettings.stiffnessConstant = 100;


            //The elbow is like a hinge, but it can't hyperflex.
            elbow.hingeLimit.isActive = true;
            elbow.hingeLimit.minimumAngle = 0;
            elbow.hingeLimit.maximumAngle = MathHelper.Pi * .7f;
            space.add(elbow);

            Entity hand = new Box(lowerArm.centerPosition + new Vector3(0, .9f, 0), .4f, .55f, .25f, 3);
            space.add(hand);

            ballSocketJoint = new BallSocketJoint(lowerArm, hand, lowerArm.centerPosition + new Vector3(0, .7f, 0));
            space.add(ballSocketJoint);

            //Wrists can use an ellipse limit too.
            swingLimit = new EllipseSwingLimit(lowerArm, hand, Vector3.Up, MathHelper.PiOver4, MathHelper.PiOver2);
            space.add(swingLimit);

            //Allow a little extra twist beyond the forearm.
            twistLimit = new TwistLimit(lowerArm, hand, Vector3.Up, Vector3.Up, -MathHelper.PiOver4 / 2, MathHelper.PiOver4 / 2);
            twistLimit.springSettings.stiffnessConstant = 100;
            twistLimit.springSettings.dampingConstant = 100;
            space.add(twistLimit);

            //The hand is pretty floppy without some damping.
            AngularMotor angularMotor = new AngularMotor(lowerArm, hand);
            angularMotor.settings.velocityMotor.softness = .5f;
            space.add(angularMotor);

            //Make sure the parts of the arm don't collide.
            torso.collisionRules.specificEntities.Add(upperArm, CollisionRule.noPair);
            lowerArm.collisionRules.specificEntities.Add(upperArm, CollisionRule.noPair);
            lowerArm.collisionRules.specificEntities.Add(hand, CollisionRule.noPair);
            //Since the entities were added to the space before changing the collision rules,
            //and therefore have collision pairs with each other already, we need to force a 
            //collision rule recalculation.
            //If the entities were added to the space after collision rules were set or if they
            //weren't colliding, this wouldn't be needed.
            torso.forceCollisionRuleRecalculation();
            lowerArm.forceCollisionRuleRecalculation();


            game.camera.position = new Vector3(0, 4, 20);
        }

    }
}
