/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */



#region Using Statements
using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using BEPUphysics;
using System.Diagnostics;
using BEPUphysics.Entities;
using BEPUphysics.Constraints;
using BEPUphysics.ForceFields;
using BEPUphysics.DataStructures;
using System.Threading;
using BEPUphysicsDemos.SampleCode;
using BEPUphysics.BroadPhases;
using BEPUphysics.Vehicle;
using BEPUphysicsDrawer.Models;
using BEPUphysicsDrawer.Lines;
using System.Text;
using BEPUphysicsDrawer.Font;
using BEPUphysicsDemos.Demos;
using BEPUphysics.ResourceManagement;

#endregion

namespace BEPUphysicsDemos
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class DemosGame : Game
    {

        public Camera camera;

        int currentSimulationIndex = 1;
        Demo currentSimulation;

        //Rendering Variables
        public GraphicsDeviceManager graphics;
        List<Contact> contacts = new List<Contact>();
        List<Constraint> constraints = new List<Constraint>();
        List<VertexPositionColor> collisionNormalList = new List<VertexPositionColor>();
        List<VertexPositionColor> jointLines = new List<VertexPositionColor>();
        List<VertexPositionColor> boundingBoxLines = new List<VertexPositionColor>();

        //Rendering tools
        public ModelDrawer modelDrawer;
        public LineDrawer constraintDrawer;
        public BasicEffect lineDrawer;
        public SpriteBatch UIDrawer;
        public TextDrawer dataTextDrawer;
        public TextDrawer tinyTextDrawer;

        //Content
        SpriteFont dataFont;
        SpriteFont tinyFont;
        Texture2D controlsMenu;

        //FPS calculation variables
        double FPSlastTime;
        double FPStotalSinceLast = 0;
        double FPStoDisplay;
        double averagePhysicsTime = 0;
        int FPStotalFramesSinceLast = 0;

        //Input
        public KeyboardState keyboardInput;
        public KeyboardState previousKeyboardInput;
        public GamePadState gamePadInput;
        public GamePadState previousGamePadInput;
#if WINDOWS
        public MouseState mouseInput;
#endif

        //Display Booleans        
        bool displayEntities = true;
        bool displayUI = true;
        bool displayConstraints = true;
        bool displayMenu = false;
        bool displayContacts = false;
        bool displayBoundingBoxes = false;
        bool displaySimulationIslands = false;

        Type[] demoTypes = new Type[]
        {
            typeof(WallDemo),
            typeof(IncomingDemo),
            typeof(VariousThingsDemo),
            typeof(TerrainDemo),
            typeof(LotsOSpheresDemo),
            typeof(MultipendulumDemo),
            typeof(SawContraptionDemo),
            typeof(PlinkoDemo),
            typeof(DogbotDemo),
            typeof(ActionFigureDemo),
            typeof(CoefficientsDemo),
            typeof(ColosseumDemo),
            typeof(CollisionFilteringDemo),
            typeof(UnfortunateGuyDemo),
            typeof(RobotArmDemo),
            typeof(PyramidDemo),
            typeof(CompoundBodiesDemo),
            typeof(SpaceshipDemo),
            typeof(JengaDemo),
            typeof(SleepModeDemo),
            typeof(BridgeDemo),
            typeof(ClothDemo),
            typeof(BroadPhaseDemo),
            typeof(PathFollowingDemo),
            typeof(BuoyancyDemo),
            typeof(TornadoDemo),
            typeof(GravitationalFieldDemo),
            typeof(PlaygroundDemo),
            typeof(FancyShapesDemo),
            typeof(FishInABarrelDemo),
            typeof(EarthquakeDemo),
            typeof(DetectorVolumeDemo),
        };



        public DemosGame()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            graphics.PreferredBackBufferWidth = 800;
            graphics.PreferredBackBufferHeight = 600;
            camera = new Camera(Vector3.Zero, 10, 0, 0, Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4, 800 / 600f, .1f, 10000));


            Exiting += DemosGame_Exiting;

        }

        void DemosGame_Exiting(object sender, EventArgs e)
        {
            currentSimulation.cleanUp();
        }


        protected override void Initialize()
        {

            if (GraphicsDevice.GraphicsProfile == GraphicsProfile.HiDef)
                modelDrawer = new InstancedModelDrawer(this);
            else
                modelDrawer = new BruteModelDrawer(this);

            constraintDrawer = new LineDrawer(this);
            constraintDrawer.displayTypes.Add(typeof(GrabSpring), typeof(DisplayGrabSpring));

            base.Initialize();
        }


        /// <summary>
        /// Manages the switch to a new physics engine simulation.
        /// </summary>
        /// <param name="sim">Index of the simulation to switch to.</param>
        void switchSimulation(int sim)
        {
            currentSimulationIndex = sim;

            //Clear out any old rendering stuff.
            modelDrawer.clear();
            constraintDrawer.clear();

            //Tell the previous simulation it's done.
            if (currentSimulation != null)
            {
                currentSimulation.cleanUp();
            }
            //Create the new demo.
            Type demoType = demoTypes[currentSimulationIndex - 1];

#if !WINDOWS
            currentSimulation = (Demo)demoType.GetConstructor(
                                                    new Type[] { typeof(DemosGame) })
                                                    .Invoke(new object[] { this });
#else
            currentSimulation = (Demo)Activator.CreateInstance(demoType, new object[] { this });
#endif

            #region DisplayObject creation
            foreach (Entity e in currentSimulation.space.entities)
            {
                if ((string)e.tag != "noDisplayObject")
                {
                    modelDrawer.add(e);
                }
                else//Remove the now unnecessary tag.
                    e.tag = null;
            }
            foreach (SolverUpdateable s in currentSimulation.space.solverUpdateables)
            {
                //Add the solver updateable and match up the activity setting.
                LineDisplayObjectBase objectAdded = constraintDrawer.add(s);
                if (objectAdded != null)
                    objectAdded.isActive = s.isActive;
            }
            #endregion

            GC.Collect();
        }

        protected override void LoadContent()
        {
            dataFont = Content.Load<SpriteFont>("DataFont");
            tinyFont = Content.Load<SpriteFont>("TinyFont");

            controlsMenu = Content.Load<Texture2D>("bepuphysicscontrols");

            IsFixedTimeStep = false;

            lineDrawer = new BasicEffect(GraphicsDevice);

            UIDrawer = new SpriteBatch(GraphicsDevice);

            dataTextDrawer = new TextDrawer(UIDrawer, dataFont, Color.White);
            tinyTextDrawer = new TextDrawer(UIDrawer, tinyFont, Color.White);

#if WINDOWS
            Mouse.SetPosition(200, 200); //This helps the camera stay on track even if the mouse is offset during startup.
#endif

            switchSimulation(1);
        }




        protected override void UnloadContent()
        {
            base.UnloadContent();
        }

        /// <summary>
        /// Determines whether or not the key was pressed this frame.
        /// </summary>
        /// <param name="key">Key to check.</param>
        /// <returns>Whether or not the key was pressed.</returns>
        public bool wasKeyPressed(Keys key)
        {
            return keyboardInput.IsKeyDown(key) && previousKeyboardInput.IsKeyUp(key);
        }

        /// <summary>
        /// Determines whether or not the button was pressed this frame.
        /// </summary>
        /// <param name="button">Button to check.</param>
        /// <returns>Whether or not the button was pressed.</returns>
        public bool wasButtonPressed(Buttons button)
        {
            return gamePadInput.IsButtonDown(button) && previousGamePadInput.IsButtonUp(button);
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {


            previousKeyboardInput = keyboardInput;
            keyboardInput = Keyboard.GetState();
            float dt = (float)gameTime.ElapsedGameTime.TotalSeconds;
#if WINDOWS
            mouseInput = Mouse.GetState();

            //Keep the mouse within the screen
            Mouse.SetPosition(200, 200);
#endif
            previousGamePadInput = gamePadInput;
            for (int i = 0; i < 4; i++)
            {
                gamePadInput = GamePad.GetState((PlayerIndex)i);
                if (gamePadInput.IsConnected)
                    break;
            }

            // Allows the default game to exit on Xbox 360 and Windows
            if (keyboardInput.IsKeyDown(Keys.Escape) || gamePadInput.Buttons.Back == ButtonState.Pressed)
                this.Exit();


            #region Camera
            //Update the camera
#if !WINDOWS

            camera.update(dt, keyboardInput, gamePadInput);
#else
            camera.update(dt, keyboardInput, mouseInput, gamePadInput);
#endif
            #endregion

            #region UI Toggles

#if !WINDOWS
            if (wasButtonPressed(Buttons.Start))
            {
                displayMenu = !displayMenu;
            }
#else
            if (wasKeyPressed(Keys.F1))
            {
                displayMenu = !displayMenu;
            }
#endif
            if (wasKeyPressed(Keys.I))
            {
                displayUI = !displayUI;
            }
            if (wasKeyPressed(Keys.J))
            {
                displayConstraints = !displayConstraints;
            }
            if (wasKeyPressed(Keys.K))
            {
                displayContacts = !displayContacts;
            }
            if (wasKeyPressed(Keys.U))
            {
                displayBoundingBoxes = !displayBoundingBoxes;
            }
            if (wasKeyPressed(Keys.Y))
            {
                displayEntities = !displayEntities;
            }
            if (wasKeyPressed(Keys.H))
            {
                displaySimulationIslands = !displaySimulationIslands;
            }
            if (wasKeyPressed(Keys.G))
            {
                modelDrawer.isWireframe = !modelDrawer.isWireframe;
            }
            #endregion
            #region Simulation Switcharoo
#if !WINDOWS

            int switchTo = -2;
            if (wasButtonPressed(Buttons.DPadLeft))
                switchTo = currentSimulationIndex - 1;
            else if (wasButtonPressed(Buttons.DPadRight))
                switchTo = currentSimulationIndex + 1;
            if (switchTo != -2)
            {
                if (switchTo < 1)
                    switchTo += demoTypes.Length;
                else if (switchTo > demoTypes.Length)
                    switchTo = 1;
                switchSimulation(switchTo);
            }
#else

            int code;
            int simNumber;
            foreach (Keys key in keyboardInput.GetPressedKeys())
            {
                code = key.GetHashCode();

                if (code >= 48 && code <= 57)
                {
                    if (code == 48)
                        simNumber = 10;
                    else
                        simNumber = code - 48;
                    if (keyboardInput.IsKeyDown(Keys.LeftShift))
                        simNumber += 10;
                    if (keyboardInput.IsKeyDown(Keys.LeftControl))
                        simNumber += 20;

                    if (simNumber <= demoTypes.Length)
                    {
                        (currentSimulation.space.threadManager as IDisposable).Dispose(); //While it would clean up by itself, this gets it done a little quicker.
                        switchSimulation(simNumber);
                    }
                }
            }



#endif
            #endregion

            currentSimulation.update(dt);

            constraintDrawer.update();

            modelDrawer.update();
            base.Update(gameTime);


        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {

            GraphicsDevice.Clear(new Color(.41f, .41f, .45f, 1));

            if (displayEntities)
                modelDrawer.draw(camera.viewMatrix, camera.projectionMatrix);

            if (displayConstraints)
                constraintDrawer.draw(camera.viewMatrix, camera.projectionMatrix);

            lineDrawer.LightingEnabled = false;
            lineDrawer.VertexColorEnabled = true;
            lineDrawer.World = Matrix.Identity;
            lineDrawer.View = camera.viewMatrix;
            lineDrawer.Projection = camera.projectionMatrix;

            #region Contact Drawing
            //Draw collisions.
            if (displayContacts)
            {

                contacts.Clear();
                foreach (CollisionPair pair in currentSimulation.space.collisionPairs)
                    contacts.AddRange(pair.contacts);
                if (contacts.Count > 0)
                {
                    collisionNormalList.Clear();
                    for (int i = 0; i < contacts.Count; i++)
                    {
                        collisionNormalList.Add(new VertexPositionColor(contacts[i].position, Color.White));
                        collisionNormalList.Add(new VertexPositionColor(contacts[i].position + contacts[i].normal, Color.White));
                    }

                    foreach (EffectPass pass in lineDrawer.CurrentTechnique.Passes)
                    {
                        pass.Apply();

                        GraphicsDevice.DrawUserPrimitives(PrimitiveType.LineList, collisionNormalList.ToArray(), 0, contacts.Count);

                    }

                }
            }
            #endregion
            #region bounding box drawing
            if (displayBoundingBoxes)
            {

                if (currentSimulation.space.entities.Count > 0)
                {
                    boundingBoxLines.Clear();
                    Vector3[] boundingBoxCorners;

                    foreach (Entity e in currentSimulation.space.entities)
                    {
                        boundingBoxCorners = e.boundingBox.GetCorners();
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[0], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[1], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[0], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[3], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[0], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[4], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[1], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[2], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[1], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[5], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[2], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[3], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[2], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[6], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[3], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[7], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[4], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[5], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[4], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[7], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[5], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[6], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[6], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[7], Color.DarkRed));

                    }
                    foreach (EffectPass pass in lineDrawer.CurrentTechnique.Passes)
                    {
                        pass.Apply();
                        GraphicsDevice.DrawUserPrimitives(PrimitiveType.LineList, boundingBoxLines.ToArray(), 0, currentSimulation.space.entities.Count * 12);

                    }

                }
            }
            #endregion
            #region Simulation Island Drawing
            if (displaySimulationIslands)
            {

                if (currentSimulation.space.entities.Count > 0)
                {
                    boundingBoxLines.Clear();
                    Vector3[] boundingBoxCorners;
                    BoundingBox box;
                    Color colorToUse;

                    foreach (SimulationIsland s in currentSimulation.space.simulationIslands)
                    {
                        List<Entity> entityList = Resources.getEntityList();
#if !WINDOWS
                        entityList.AddRange(s.entities.Keys);
#else
                        entityList.AddRange(s.entities);
#endif

                        box = entityList[0].boundingBox;
                        for (int k = 1; k < entityList.Count; k++)
                        {
                            box = BoundingBox.CreateMerged(box, entityList[k].boundingBox);
                        }
                        Resources.giveBack(entityList);

                        if (s.isCurrentlyActive)
                            colorToUse = Color.DarkGoldenrod;
                        else
                            colorToUse = Color.DarkGray;
                        boundingBoxCorners = box.GetCorners();
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[0], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[1], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[0], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[3], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[0], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[4], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[1], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[2], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[1], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[5], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[2], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[3], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[2], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[6], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[3], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[7], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[4], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[5], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[4], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[7], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[5], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[6], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[6], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[7], colorToUse));

                    }

                    if (currentSimulation.space.simulationIslands.Count > 0)
                        foreach (EffectPass pass in lineDrawer.CurrentTechnique.Passes)
                        {
                            pass.Apply();
                            GraphicsDevice.DrawUserPrimitives(PrimitiveType.LineList, boundingBoxLines.ToArray(), 0, currentSimulation.space.simulationIslands.Count * 12);

                        }

                }
            }
            #endregion
            #region UI Drawing
            UIDrawer.Begin();
            if (displayUI)
            {
                FPStotalSinceLast += gameTime.ElapsedGameTime.TotalSeconds;
                FPStotalFramesSinceLast++;
                if (gameTime.TotalGameTime.TotalSeconds - FPSlastTime > .25 && gameTime.ElapsedGameTime.TotalSeconds > 0)
                {
                    double avg = FPStotalSinceLast / FPStotalFramesSinceLast;
                    FPSlastTime = gameTime.TotalGameTime.TotalSeconds;
                    FPStoDisplay = Math.Round(1 / avg, 1);
                    averagePhysicsTime = Math.Round(1000 * currentSimulation.physicsTime, 1);
                    FPStotalSinceLast = 0;
                    FPStotalFramesSinceLast = 0;

                }

                dataTextDrawer.draw("FPS: ", FPStoDisplay, new Vector2(50, 450));
                dataTextDrawer.draw("Physics Time (ms): ", averagePhysicsTime, new Vector2(50, 467));
                dataTextDrawer.draw("Collision Pairs: ", currentSimulation.space.collisionPairs.Count, new Vector2(50, 484));
                int countActive = 0;

                for (int i = 0; i < currentSimulation.space.entities.Count; i++)
                {
                    if (currentSimulation.space.entities[i].isActive)
                        countActive++;
                }
                dataTextDrawer.draw("Active Objects: ", countActive, new Vector2(50, 501));
#if !WINDOWS
                dataTextDrawer.draw("Press Start for Controls", new Vector2(50, 518));
#else
                dataTextDrawer.draw("Press F1 for Controls", new Vector2(50, 518));
#endif

                tinyTextDrawer.draw("Current Simulation: ", currentSimulationIndex, new Vector2(600, 500));
                tinyTextDrawer.draw(currentSimulation.name, new Vector2(620, 514));

                currentSimulation.drawUI();
            }
            if (displayMenu)
            {
                UIDrawer.Draw(controlsMenu, new Rectangle(0, 0, 800, 600), Color.White);
            }
            UIDrawer.End();


            #endregion

            base.Draw(gameTime);

        }


    }
}