/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using BEPUphysics.Entities;
using BEPUphysics;

namespace BEPUphysicsDemos.SampleCode
{

    /// <summary>
    /// Handles radial impulse applications on nearby objects when activated.
    /// </summary>
    public class Explosion
    {
        /// <summary>
        /// Gets or sets the current position of the explosion.
        /// </summary>
        public Vector3 position { get; set; }
        /// <summary>
        /// Gets or sets the base strength of the blast.
        /// </summary>
        public float magnitude { get; set; }

        /// <summary>
        /// Gets or sets the maximum distance that the explosion will affect.
        /// </summary>
        public float maxDistance { get; set; }

        /// <summary>
        /// Gets or sets the space that the explosion will explode in.
        /// </summary>
        public Space space { get; set; }

        /// <summary>
        /// Re-used list of entities hit by the explosion.
        /// </summary>
        List<Entity> affectedEntities = new List<Entity>();

        /// <summary>
        /// Constructs an explosion.
        /// </summary>
        /// <param name="pos">Initial position of the explosion.</param>
        /// <param name="explosionMagnitude">Base strength of the blast as applied in units of impulse.</param>
        /// <param name="maxDist">Maximum radius of effect.</param>
        /// <param name="containingSpace">Space in which the explosion resides.</param>
        public Explosion(Vector3 pos, float explosionMagnitude, float maxDist, Space containingSpace)
        {
            position = pos;
            magnitude = explosionMagnitude;
            maxDistance = maxDist;
            space = containingSpace;
        }

        /// <summary>
        /// Detonates the explosion, applying impulses to applicable physically simulated entities.
        /// </summary>
        public void explode()
        {
            affectedEntities.Clear();
            //Ask the broadphase system which entities are in the explosion region.
            space.broadPhase.getEntities(new BoundingSphere(position, maxDistance), affectedEntities);

            foreach (Entity e in affectedEntities)
            {
                //Don't bother applying impulses to kinematic entities; they have infinite inertia.
                if (e.isDynamic)
                {
                    Vector3 offset = e.internalCenterPosition - position;
                    float distanceSquared = offset.LengthSquared();
                    if (distanceSquared > Toolbox.epsilon) //Be kind to the engine and don't give it a value divided by zero.
                    {
                        float distance = (float)Math.Sqrt(distanceSquared);
                        //This applies a force inversely proportional to the distance.
                        //Note the extra distance term in the denominator.  This normalizes the
                        //offset, resulting in a quadratic explosion falloff.
                        //A linear falloff could be accomplished by not including the extra distance term.
                        e.applyLinearImpulse(offset * (magnitude / (distanceSquared * distance)));
                        //The above only applies a linear impulse, which is quick and usually sufficient to look like an explosion.
                        //If you want some extra chaotic spinning, try applying an angular impulse.
                        //Using e.applyImpulse with an appropriate impulse location or e.applyAngularImpulse will do the job.

                        //Since applyLinearImpulse doesn't wake up an entity by default, force the exploded entity awake.
                        e.isActive = true;
                    }
                    else
                    {
                        e.applyLinearImpulse(new Vector3(0, magnitude, 0));
                    }
                }
            }
        }



    }
}
