/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using BEPUphysics.Entities;
using BEPUphysics.ForceFields;

namespace BEPUphysicsDemos.SampleCode
{
    /// <summary>
    /// Applies a force on all objects within a field proportional to the inverse square of the distance.
    /// </summary>
    public class GravitationalField : ForceField
    {
        /// <summary>
        /// Gets or sets the gravitational constant of the field times the effective mass at the center of the field.
        /// </summary>
        public float multiplier { get; set; }

        /// <summary>
        /// Gets or sets the maximum force that can be applied by the field.
        /// </summary>
        public float maxForce { get; set; }

        /// <summary>
        /// Gets or sets the center of the field that entities will be pushed toward.
        /// </summary>
        public Vector3 origin { get; set; }

        /// <summary>
        /// Creates a gravitational field.
        /// </summary>
        /// <param name="shape">Shape representing the volume of the force field.</param>
        /// <param name="origin">Location that entities will be pushed toward.</param>
        /// <param name="multiplier">Represents the gravitational constant of the field times the effective mass at the center of the field.</param>
        /// <param name="maxForce">Maximum force the field can apply.</param>
        public GravitationalField(ForceFieldShape shape, Vector3 origin, float multiplier, float maxForce)
            : base(shape)
        {
            this.multiplier = multiplier;
            this.origin = origin;
            this.maxForce = maxForce;
        }


        /// <summary>
        /// Calculates the gravitational force to apply to the entity.
        /// </summary>
        /// <param name="e">Target of the impulse.</param>
        /// <param name="dt">Time since the last frame in simulation seconds.</param>
        /// <param name="impulse">Force to apply at the given position.</param>
        protected override void calculateImpulse(Entity e, float dt, out Vector3 impulse)
        {
            //Could use a linear dropoff for a slightly faster calculation (divide by lengthSquared instead of length^3)
            //The third length power here is to normalize the direction r.
            Vector3 r = e.internalCenterPosition - origin;
            float length = r.Length();
            float force = dt * Math.Min(maxForce, multiplier * e.mass / (length * length * length));
            impulse = -force * r;

        }
    }
}
