/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using System.Diagnostics;
using BEPUphysics.Entities;
using BEPUphysics.ForceFields;
using BEPUphysics;

namespace BEPUphysicsDemos.SampleCode
{
    /// <summary>
    /// Force field simulating a tornado with circular a 'wind' force, an inward pointing force, and an upward pointing force.
    /// </summary>
    public class Tornado : ForceField
    {
        /// <summary>
        /// Gets or sets the position of the tornado.  This is only the origin of the force; move the shape along with it if the position moves away from the shape.
        /// </summary>
        public Vector3 position { get; set; }
        /// <summary>
        /// Maximum horizontal wind speed of the tornado; objects will not be accelerated by the wind past this speed in the direction of the wind.
        /// </summary>
        public float horizontalWindSpeed;
        /// <summary>
        /// Maximum upward pushing wind speed; objects will not be accelerated by the wind past this speed upward.
        /// </summary>
        public float upwardSuctionSpeed;
        /// <summary>
        /// Maximum inward sucking wind speed; objects will not be accelerated by the wind past this speed inward.
        /// </summary>
        public float inwardSuctionSpeed;
        /// <summary>
        /// Circular force applied within the tornado.  Force magnitude decreases as distance from axis increases past the radius.
        /// </summary>
        public float horizontalForce;
        /// <summary>
        /// Magnitude of upward-pushing force within the tornado.  Magnitude decreases as distance from the axis increases past the radius.
        /// </summary>
        public float upwardForce;
        /// <summary>
        /// Magnitude of the inward-sucking force within the tornado.  Magnitude decreases as distance from the axis increases past the radius.
        /// </summary>
        public float inwardForce;
        /// <summary>
        /// Radius of the tornado at the bottom.
        /// </summary>
        public float bottomRadius;
        /// <summary>
        /// Radius of the tornado at the top.
        /// </summary>
        public float topRadius;
        /// <summary>
        /// Height of the tornado; objects above or below the tornado will not be affected by its winds.
        /// </summary>
        public float height;
        /// <summary>
        /// Axis of rotation of the tornado.
        /// </summary>
        public Vector3 axis;
        /// <summary>
        /// Spin direction of the tornado.  Looking down from the top of the tornado (the furthest forward along the tornado axis).
        /// </summary>
        public bool spinClockwise = true;

        /// <summary>
        /// Creates a simple, constant force field.
        /// </summary>
        /// <param name="shape">Shape representing the tornado-affected volume.</param>
        /// <param name="position">Position of the tornado.</param>
        /// <param name="axis">Axis of rotation of the tornado.</param>
        /// <param name="height">Height of the tornado; objects above or below the tornado will not be affected by its winds.</param>
        /// <param name="spinClockwise">Whether or not the tornado's rotation is clockwise.</param>
        /// <param name="horizontalWindSpeed">Maximum tangential wind speed; objects will not be accelerated by the wind past this speed sideways.</param>
        /// <param name="upwardSuctionSpeed">Maximum upward pushing wind speed; objects will not be accelerated by the wind past this speed upward.</param>
        /// <param name="inwardSuctionSpeed">Maximum inward sucking wind speed; objects will not be accelerated by the wind past this speed inward.</param>
        /// <param name="horizontalForce">Circular force applied within the tornado.  Force magnitude decreases as distance from axis increases past the radius.</param>
        /// <param name="upwardForce">Magnitude of upward-pushing force within the tornado.  Magnitude decreases as distance from the axis increases past the radius.</param>
        /// <param name="inwardForce">Magnitude of the inward-sucking force within the tornado.  Magnitude decreases as distance from the axis increases past the radius.</param>
        /// <param name="topRadius">Radius of the tornado at the top.</param>
        /// <param name="bottomRadius">Radius of the tornado at the bottom.</param>
        public Tornado(ForceFieldShape shape, Vector3 position, Vector3 axis,
            float height, bool spinClockwise, float horizontalWindSpeed,
            float upwardSuctionSpeed, float inwardSuctionSpeed,
            float horizontalForce, float upwardForce, float inwardForce,
            float topRadius, float bottomRadius)
            : base(shape)
        {
            this.axis = Vector3.Normalize(axis);
            this.position = position;
            this.height = height;
            this.spinClockwise = spinClockwise;
            this.horizontalWindSpeed = horizontalWindSpeed;
            this.upwardSuctionSpeed = upwardSuctionSpeed;
            this.inwardSuctionSpeed = inwardSuctionSpeed;
            this.horizontalForce = horizontalForce;
            this.upwardForce = upwardForce;
            this.inwardForce = inwardForce;
            this.bottomRadius = bottomRadius;
            this.topRadius = topRadius;
        }

        /// <summary>
        /// Calculates the impulse to apply to the center of mass of physically simulated bodies within the volume.
        /// </summary>
        /// <param name="e">Target of the impulse.</param>
        /// <param name="dt">Time since the last frame in simulation seconds.</param>
        /// <param name="impulse">Force to apply at the given position.</param>
        protected override void calculateImpulse(Entity e, float dt, out Vector3 impulse)
        {
            Vector3 position = this.position; //Referenced a lot, and passed using ref parameter.
            Vector3 entityPosition = e.internalCenterPosition;

            float entityHeight = Vector3.Dot(axis, entityPosition - position + axis * (height / 2));
            if (entityHeight < 0 || entityHeight > height)
                impulse = Toolbox.zeroVector;
            else
            {
                float tornadoRadius = bottomRadius * (1 - entityHeight / height) + topRadius * (entityHeight / height);
                Vector3 closestPoint;
                Vector3 endpointA = position + axis * height / 2;
                Vector3 endpointB = position - axis * height / 2;
                Toolbox.getClosestPointOnSegmentToPoint(ref endpointA, ref endpointB, ref entityPosition, out closestPoint);
                float entityDistanceFromTornado;
                Vector3.Distance(ref entityPosition, ref closestPoint, out entityDistanceFromTornado);
                //Compute the axis to the 
                Vector3 posClosest;
                float forceMultiplier;
                if (entityDistanceFromTornado > tornadoRadius)
                {//outside tornado
                    forceMultiplier = tornadoRadius / entityDistanceFromTornado;
                    posClosest = (closestPoint - entityPosition) / entityDistanceFromTornado;
                }
                else if (entityDistanceFromTornado > Toolbox.epsilon)
                {//inside tornado
                    forceMultiplier = .5f + .5f * entityDistanceFromTornado / tornadoRadius;
                    posClosest = (closestPoint - entityPosition) / entityDistanceFromTornado;
                }
                else
                {
                    forceMultiplier = .5f;
                    posClosest = Toolbox.zeroVector;
                }

                Vector3 tangentialForceVector;
                //Don't need to normalize the direction.  
                //Axis and posClosest are perpendicular and each normal, so the result is normal.
                Vector3 tangentDirection;

                if (spinClockwise)
                    Vector3.Cross(ref axis, ref posClosest, out tangentDirection);
                else
                    Vector3.Cross(ref posClosest, ref axis, out tangentDirection);

                //Current velocity along the tangent direction.
                float dot = Vector3.Dot(e.internalLinearVelocity, tangentDirection);
                //Compute the velocity difference between the current and the maximum
                dot = horizontalWindSpeed - dot;
                //Compute the force needed to reach the maximum, but clamp it to the amount of force that the tornado can apply
                dot = MathHelper.Clamp(dot * e.mass, 0, horizontalForce * dt);
                Vector3.Multiply(ref tangentDirection, dot, out tangentialForceVector);

                //Do a similar process for the other tornado force axes.
                Vector3 upwardForceVector;
                dot = Vector3.Dot(e.internalLinearVelocity, axis);
                dot = upwardSuctionSpeed - dot;
                dot = MathHelper.Clamp(dot * e.mass, 0, upwardForce * dt);
                Vector3.Multiply(ref axis, dot, out upwardForceVector);

                Vector3 inwardForceVector;
                dot = Vector3.Dot(e.internalLinearVelocity, posClosest);
                dot = inwardSuctionSpeed - dot;
                dot = MathHelper.Clamp(dot * e.mass, 0, inwardForce * dt);
                Vector3.Multiply(ref posClosest, dot, out inwardForceVector);

                //if (posClosest.X > 0)
                //    Debug.WriteLine("Break.");
                impulse = forceMultiplier * (tangentialForceVector + upwardForceVector + inwardForceVector);
            }
        }
    }
}
