﻿/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */

using BEPUphysics;
using Microsoft.Xna.Framework;
using BEPUphysics.Entities;

namespace BEPUphysicsDemos.SampleCode
{
    /// <summary>
    /// Simple example updateable that acts like a rocket engine.
    /// </summary>
    public class Thruster : Updateable
    {
        /// <summary>
        /// Gets or sets the position of the thruster in the local space of the target entity.
        /// </summary>
        public Vector3 position { get; set; }
        /// <summary>
        /// Gets or sets the direction of the force in the local space of the target entity.  Magnitude of the force is equal to the magnitude of the direction.
        /// </summary>
        public Vector3 direction { get; set; }

        /// <summary>
        /// Gets or sets the entity to apply force to.
        /// </summary>
        public Entity target { get; set; }

        float myAge = 0;
        /// <summary>
        /// Gets or sets the length of time that the thruster has been firing.
        /// This can be reset to 'refresh' the life of the force.
        /// </summary>
        public float age
        {
            get
            {
                return myAge;
            }
            set
            {
                myAge = value;
                if (myAge < lifeSpan)
                    isUpdating = true; //isUpdating is a property of the Updateable class.  The updateDuringForces method won't be called if isUpdating is false.
            }
        }

        float myLifeSpan;
        /// <summary>
        /// Maximum life span of the force, after which the thruster will deactivate.
        /// Set to 0 for infinite lifespan.
        /// </summary>
        public float lifeSpan
        {
            get
            {
                return myLifeSpan;
            }
            set
            {
                myLifeSpan = value;
                if (myLifeSpan > age || myLifeSpan == 0)
                    isUpdating = true; //Wake the thruster up if it's young again.
            }
        }


        /// <summary>
        /// Constructs a thruster originating at the given position, pushing in the given direction.
        /// </summary>
        /// <param name="targetEntity">Entity that the force will be applied to.</param>
        /// <param name="pos">Origin of the force.</param>
        /// <param name="dir">Direction of the force.</param>
        /// <param name="time">Total lifespan of the force.  A lifespan of zero is infinite.</param>
        public Thruster(Entity targetEntity, Vector3 pos, Vector3 dir, float time)
        {
            target = targetEntity;
            position = pos;
            direction = dir;
            lifeSpan = time;
        }

        /// <summary>
        /// Applies the thruster forces.
        /// Called automatically by the owning space during a space update.
        /// </summary>
        /// <param name="dt">Simulation timestep.</param>
        public override void updateDuringForces(float dt)
        {
            //Transform the local position and direction into world space.
            Vector3 worldPosition = target.internalCenterOfMass + Vector3.TransformNormal(position, target.internalOrientationMatrix);
            Vector3 worldDirection = Vector3.TransformNormal(direction, target.internalOrientationMatrix);
            //Apply the force.
            target.applyImpulse(worldPosition, worldDirection * dt);


            age += dt;
            if (lifeSpan > 0 && age > lifeSpan)
            {
                isUpdating = false; //The thruster has finished firing.
            }
        }


    }
}
