using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;
using System.Globalization;

namespace BEPUphysicsDrawer.Font
{
    /// <summary>
    /// Simple text-drawing helper class.
    /// </summary>
    public class TextDrawer
    {
        /// <summary>
        /// Gets the sprite batch used by the drawer.
        /// </summary>
        public SpriteBatch spriteBatch { get; private set; }

        /// <summary>
        /// Gets or sets the font used by the drawer.
        /// </summary>
        public SpriteFont font { get; set; }

        /// <summary>
        /// Gets or sets the color used by the drawer.
        /// </summary>
        public Color color { get; set; }

        /// <summary>
        /// Constructs a new text drawer.
        /// </summary>
        /// <param name="spriteBatch">Sprite batch to use.</param>
        /// <param name="font">Font to use.</param>
        /// <param name="color">Color to use.</param>
        public TextDrawer(SpriteBatch spriteBatch, SpriteFont font, Color color)
        {
            this.spriteBatch = spriteBatch;
            this.font = font;
            this.color = color;
        }

        /// <summary>
        /// Draws the text at a position.
        /// </summary>
        /// <param name="text">Text to draw.</param>
        /// <param name="position">Position to draw the text at.</param>
        public void draw(string text, Vector2 position)
        {
            spriteBatch.DrawString(font, text, position, color);
        }

        /// <summary>
        /// Draws text followed by an integer.
        /// </summary>
        /// <param name="text">Text to draw.</param>
        /// <param name="value">Value to draw.</param>
        /// <param name="position">Position to draw the text at.</param>
        public void draw(string text, int value, Vector2 position)
        {
            spriteBatch.DrawString(font, text, position, color);
            draw(value, new Vector2(position.X + font.MeasureString(text).X, position.Y));
        }


        /// <summary>
        /// Draws text followed by a double.
        /// </summary>
        /// <param name="text">Text to draw.</param>
        /// <param name="value">Number to draw.</param>
        /// <param name="position">Position to draw the text at.</param>
        public void draw(string text, double value, Vector2 position)
        {
            spriteBatch.DrawString(font, text, position, color);
            draw(value, new Vector2(position.X + font.MeasureString(text).X, position.Y));
        }

        /// <summary>
        /// Draws text followed by a double.
        /// </summary>
        /// <param name="text">Text to draw.</param>
        /// <param name="value">Number to draw.</param>
        /// <param name="numPlaces">Number of digits to include after the decimal.</param>
        /// <param name="position">Position to draw the text at.</param>
        public void draw(string text, double value, int numPlaces, Vector2 position)
        {
            spriteBatch.DrawString(font, text, position, color);
            draw(value, numPlaces, new Vector2(position.X + font.MeasureString(text).X, position.Y));
        }

        //These last bits are based on some sample code posted by Stephen Styrchak (http://forums.xna.com/forums/p/45512/271953.aspx).
        private static string[] digits = { "0", "1", "2", "3", "4", "5", "6", "7", "8", "9" };
        private static string dot = ".";
        private static string[] characterBuffer = new string[310];
        private static float[] xPositionBuffer = new float[310];
        private static readonly string minValue = Int32.MinValue.ToString(CultureInfo.InvariantCulture);

        /// <summary>  
        /// Draws an integer.
        /// </summary>  
        /// <param name="value">The integer value to draw.</param>  
        /// <param name="position">The screen position specifying where to draw the value.</param>  
        /// <returns>The next position on the line to draw text. This value uses position.Y and position.X plus the equivalent of calling spriteFont.MeasureString on value.ToString(CultureInfo.InvariantCulture).</returns>  
        public Vector2 draw(int value, Vector2 position)
        {
            Vector2 nextPosition = position;

            if (value == Int32.MinValue)
            {
                nextPosition.X = nextPosition.X + font.MeasureString(minValue).X;
                spriteBatch.DrawString(font, minValue, position, color);
                position = nextPosition;
            }
            else
            {
                if (value < 0)
                {
                    nextPosition.X = nextPosition.X + font.MeasureString("-").X;
                    spriteBatch.DrawString(font, "-", position, color);
                    value = -value;
                    position = nextPosition;
                }

                int index = 0;

                do
                {
                    int modulus = value % 10;
                    value = value / 10;

                    characterBuffer[index] = digits[modulus];
                    xPositionBuffer[index] = font.MeasureString(digits[modulus]).X;
                    index += 1;
                }
                while (value > 0);

                for (int i = index - 1; i >= 0; --i)
                {
                    nextPosition.X = nextPosition.X + xPositionBuffer[i] + font.Spacing;
                    spriteBatch.DrawString(font, characterBuffer[i], position, color);
                    position = nextPosition;
                }
            }
            return position;
        }

        /// <summary>  
        /// Draws a double.
        /// </summary>  
        /// <param name="value">The double value to draw.</param>  
        /// <param name="position">The screen position specifying where to draw the value.</param>  
        /// <returns>The next position on the line to draw text. This value uses position.Y and position.X plus the equivalent of calling spriteFont.MeasureString on value.ToString(CultureInfo.InvariantCulture).</returns>  
        public Vector2 draw(double value, Vector2 position)
        {
            return draw(value, 1, position);
        }

        /// <summary>  
        /// Draws a double.
        /// </summary>  
        /// <param name="value">The double value to draw.</param>  
        /// <param name="numPlaces">Maximum number of digits to include after the decimal.</param>
        /// <param name="position">The screen position specifying where to draw the value.</param>  
        /// <returns>The next position on the line to draw text. This value uses position.Y and position.X plus the equivalent of calling spriteFont.MeasureString on value.ToString(CultureInfo.InvariantCulture).</returns>  
        public Vector2 draw(double value, int numPlaces, Vector2 position)
        {
            Vector2 nextPosition = position;

            if (value < 0)
            {
                nextPosition.X = nextPosition.X + font.MeasureString("-").X;
                spriteBatch.DrawString(font, "-", position, color);
                value = -value;
                position = nextPosition;
            }

            int index = 0;
            value = Math.Abs(value);

            double wholeValue = value;
            value = value % 1;
            if (value != 0)
            {
                value *= Math.Pow(10, numPlaces);

                do
                {
                    long modulus = (long)value % 10;
                    value = (long)value / 10;

                    characterBuffer[index] = digits[modulus];
                    xPositionBuffer[index] = font.MeasureString(digits[modulus]).X;
                    index += 1;
                }
                while (value > 0 || index < numPlaces);

                characterBuffer[index] = dot;
                xPositionBuffer[index] = font.MeasureString(dot).X;
                index += 1;
            }

            do
            {
                long modulus = (long)wholeValue % 10;
                wholeValue = (long)wholeValue / 10;

                characterBuffer[index] = digits[modulus];
                xPositionBuffer[index] = font.MeasureString(digits[modulus]).X;
                index += 1;
            }
            while (wholeValue > 0);



            for (int i = index - 1; i >= 0; --i)
            {
                nextPosition.X = nextPosition.X + xPositionBuffer[i] + font.Spacing;
                spriteBatch.DrawString(font, characterBuffer[i], position, color);
                position = nextPosition;
            }

            return position;
        }
    }

}