﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Collections.ObjectModel;

namespace BEPUphysicsDrawer.Models
{
    /// <summary>
    /// Manages batching of display models and their drawing.
    /// </summary>
    public class ModelDisplayObjectBatch
    {
        /// <summary>
        /// Maximum number of display objects that can be lumped into a single display batch.
        /// </summary>
        public const int maximumObjectsPerBatch = 45;
        /// <summary>
        /// Maximum number of primitives that can be batched together in a single draw call.
        /// </summary>
        public static int maximumPrimitiveCountPerBatch = 100000;

        /// <summary>
        /// List of all world transforms associated with display objects in the batch.
        /// </summary>
        Matrix[] worldTransforms = new Matrix[maximumObjectsPerBatch];
        /// <summary>
        /// List of textures associated with display objects in the batch.
        /// </summary>
        int[] textureIndices = new int[maximumObjectsPerBatch];

        VertexPositionNormalTexture[] vertices = new VertexPositionNormalTexture[0];
        ushort[] indices = new ushort[0];

        //These lists are used to collect temporary data from display objects.
        List<VertexPositionNormalTexture> vertexList = new List<VertexPositionNormalTexture>();
        List<ushort> indexList = new List<ushort>();

        VertexBuffer vertexBuffer;
        IndexBuffer indexBuffer;

        float instanceIndex;

        float[] instancingIndices = new float[0];
        /// <summary>
        /// Contains instancing data to be fed into the second stream.
        /// </summary>
        VertexBuffer instancingBuffer;

        //Has VertexBuffer, IndexBuffer, second stream for indices.
        //Why second stream?  Creating it will be easier since can addRange on lists gotten from ModelDisplayObject, then toss in
        //the appropriate indices separately.

        List<ModelDisplayObjectBase> myDisplayObjects = new List<ModelDisplayObjectBase>();
        ReadOnlyCollection<ModelDisplayObjectBase> myDisplayObjectsReadOnly;
        /// <summary>
        /// Gets the display objects in this batch.
        /// </summary>
        public ReadOnlyCollection<ModelDisplayObjectBase> displayObjects
        {
            get
            {
                return myDisplayObjectsReadOnly;
            }
        }

        GraphicsDevice graphicsDevice;


        /// <summary>
        /// Constructs a new batch.
        /// </summary>
        /// <param name="graphicsDevice">Device to use when drawing.</param>
        public ModelDisplayObjectBatch(GraphicsDevice graphicsDevice)
        {
            this.graphicsDevice = graphicsDevice;
            myDisplayObjectsReadOnly = new ReadOnlyCollection<ModelDisplayObjectBase>(myDisplayObjects);


        }

        /// <summary>
        /// Adds a display object to the batch.
        /// </summary>
        /// <param name="displayObject">Display object to add.</param>
        public bool add(ModelDisplayObjectBase displayObject, InstancedModelDrawer drawer)
        {
            //In theory, don't need to test for dupliate entries since batch.add
            //should only be called through a InstancedModelDrawer's add (which checks beforehand).
            if (myDisplayObjects.Count == maximumObjectsPerBatch ||
                ((indices.Length / 3 + displayObject.getTriangleCountEstimate()) > maximumPrimitiveCountPerBatch &&
                myDisplayObjects.Count > 0))
                return false;
            myDisplayObjects.Add(displayObject);
            int instanceIndex = myDisplayObjects.Count - 1;
            displayObject.getVertexData(vertexList, indexList, this, (ushort)vertices.Length, indices.Length, instanceIndex);
            //Add the data to the batch.
            VertexPositionNormalTexture[] newVertices = new VertexPositionNormalTexture[vertices.Length + vertexList.Count];
            vertices.CopyTo(newVertices, 0);
            vertexList.CopyTo(newVertices, vertices.Length);
            vertices = newVertices;

            ushort[] newIndices = new ushort[indices.Length + indexList.Count];
            indices.CopyTo(newIndices, 0);
            indexList.CopyTo(newIndices, indices.Length);
            indices = newIndices;

            float[] newInstancingIndices = new float[instancingIndices.Length + vertexList.Count];
            instancingIndices.CopyTo(newInstancingIndices, 0);
            for (int i = instancingIndices.Length; i < newInstancingIndices.Length; i++)
                newInstancingIndices[i] = instanceIndex;
            instancingIndices = newInstancingIndices;

            vertexBuffer = new VertexBuffer(graphicsDevice, VertexPositionNormalTexture.VertexDeclaration, vertices.Length, BufferUsage.WriteOnly);
            vertexBuffer.SetData(vertices);
            instancingBuffer = new VertexBuffer(graphicsDevice, drawer.instancingVertexDeclaration, instancingIndices.Length, BufferUsage.WriteOnly);
            instancingBuffer.SetData(instancingIndices);
            indexBuffer = new IndexBuffer(graphicsDevice, IndexElementSize.SixteenBits, indices.Length, BufferUsage.WriteOnly);
            indexBuffer.SetData(indices);

            vertexList.Clear();
            indexList.Clear();
            return true;
        }

        /// <summary>
        /// Removes a display object from the batch.
        /// </summary>
        /// <param name="displayObject">Display object to remove.</param>
        /// <param name="drawer">Instanced model drawer doing the removal.</param>
        public void remove(ModelDisplayObjectBase displayObject, InstancedModelDrawer drawer)
        {
            //Modify vertex buffer
            VertexPositionNormalTexture[] newVertices = new VertexPositionNormalTexture[vertices.Length - displayObject.batchInformation.vertexCount];
            //Copy the first part back (before the display object)
            Array.Copy(vertices, 0, newVertices, 0, displayObject.batchInformation.baseVertexBufferIndex);
            //Copy the second part back (after the display object)
            Array.Copy(vertices, displayObject.batchInformation.baseVertexBufferIndex + displayObject.batchInformation.vertexCount,
                       newVertices, displayObject.batchInformation.baseVertexBufferIndex,
                       vertices.Length - (displayObject.batchInformation.baseVertexBufferIndex + displayObject.batchInformation.vertexCount));
            vertices = newVertices;
            
            //Modify index buffer
            ushort[] newIndices = new ushort[indices.Length - displayObject.batchInformation.indexCount];
            //Copy the first part back (before the display object)
            Array.Copy(indices, 0, newIndices, 0, displayObject.batchInformation.baseIndexBufferIndex);
            //Copy the second part back (after the display object)
            Array.Copy(indices, displayObject.batchInformation.baseIndexBufferIndex + displayObject.batchInformation.indexCount,
                       newIndices, displayObject.batchInformation.baseIndexBufferIndex,
                       indices.Length - (displayObject.batchInformation.baseIndexBufferIndex + displayObject.batchInformation.indexCount));
            indices = newIndices;
            //The index buffer's data is now wrong though.  We deleted a bunch of vertices.
            //So go through the index buffer starting at the point of deletion and decrease the values appropriately.
            for (int i = displayObject.batchInformation.baseIndexBufferIndex; i < indices.Length; i++)
                indices[i] -= (ushort)displayObject.batchInformation.vertexCount;

            //Modify instancing indices
            float[] newInstancingIndices = new float[instancingIndices.Length - displayObject.batchInformation.vertexCount];
            //Copy the first part back (before the display object)
            Array.Copy(instancingIndices, 0, newInstancingIndices, 0, displayObject.batchInformation.baseVertexBufferIndex);
            //Copy the second part back (after the display object)
            Array.Copy(instancingIndices, displayObject.batchInformation.baseVertexBufferIndex + displayObject.batchInformation.vertexCount,
                       newInstancingIndices, displayObject.batchInformation.baseVertexBufferIndex,
                       instancingIndices.Length - (displayObject.batchInformation.baseVertexBufferIndex + displayObject.batchInformation.vertexCount));
            instancingIndices = newInstancingIndices;
            //Like with the index buffer, go through the buffer starting at the point of deletion and decrease the values appropriately.
            for (int i = displayObject.batchInformation.baseVertexBufferIndex; i < instancingIndices.Length; i++)
                instancingIndices[i]--;

            myDisplayObjects.Remove(displayObject);
            // Tell ModelDisplayObject that it got removed (batch, indices).
            displayObject.clearBatchReferences();

            vertexBuffer = new VertexBuffer(graphicsDevice, VertexPositionNormalTexture.VertexDeclaration, vertices.Length, BufferUsage.WriteOnly);
            vertexBuffer.SetData(vertices);
            instancingBuffer = new VertexBuffer(graphicsDevice, drawer.instancingVertexDeclaration, instancingIndices.Length, BufferUsage.WriteOnly);
            instancingBuffer.SetData(instancingIndices);
            indexBuffer = new IndexBuffer(graphicsDevice, IndexElementSize.SixteenBits, indices.Length, BufferUsage.WriteOnly);
            indexBuffer.SetData(indices);


        }

        /// <summary>
        /// Updates the display objects within the batch and collects transforms.
        /// </summary>
        public void update()
        {
            for (int i = 0; i < myDisplayObjects.Count; i++)
            {
                myDisplayObjects[i].update();
                worldTransforms[i] = myDisplayObjects[i].worldTransform;
                textureIndices[i] = myDisplayObjects[i].textureIndex;
            }
        }

        /// <summary>
        /// Draws the models managed by the batch.
        /// </summary>
        public void draw(Effect effect, EffectParameter worldTransformsParameter, EffectParameter textureIndicesParameter, EffectPass pass)
        {
            if (vertices.Length > 0)
            {
                graphicsDevice.SetVertexBuffers(vertexBuffer, instancingBuffer);
                graphicsDevice.Indices = indexBuffer;
                worldTransformsParameter.SetValue(worldTransforms);
                textureIndicesParameter.SetValue(textureIndices);
                pass.Apply();

                graphicsDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList,
                                                         0, 0, vertices.Length,
                                                         0, indices.Length / 3);
            }
        }

    }
}
