﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using BEPUphysics.Entities;
using BEPUphysics;
using BEPUphysics.DataStructures;

namespace BEPUphysicsDrawer.Models
{
    /// <summary>
    /// Manages and draws models.
    /// </summary>
    public abstract class ModelDrawer
    {
        List<SelfDrawingModelDisplayObject> selfDrawingDisplayObjects = new List<SelfDrawingModelDisplayObject>();

        Dictionary<object, ModelDisplayObjectBase> displayObjects = new Dictionary<object, ModelDisplayObjectBase>();

        protected Texture2D[] textures = new Texture2D[8];

        Dictionary<Type, Type> myDisplayTypes = new Dictionary<Type, Type>();
        /// <summary>
        /// Gets the map from object types to display object types.
        /// </summary>
        public Dictionary<Type, Type> displayTypes
        {
            get
            {
                return myDisplayTypes;
            }
        }

        /// <summary>
        /// Gets the game using this ModelDrawer.
        /// </summary>
        public Game game { get; private set; }

        RasterizerState wireframeState;
        RasterizerState fillState;

        bool myIsWireframe;
        /// <summary>
        /// Gets or sets whether or not the model drawer is drawing wireframes.
        /// </summary>
        public bool isWireframe
        {
            get
            {
                return myIsWireframe;
            }
            set
            {
                myIsWireframe = value;
            }
        }

        protected ModelDrawer(Game game)
        {
            this.game = game;
            textures[0] = game.Content.Load<Texture2D>("Textures//red");
            textures[1] = game.Content.Load<Texture2D>("Textures//yellow");
            textures[2] = game.Content.Load<Texture2D>("Textures//purple");
            textures[3] = game.Content.Load<Texture2D>("Textures//orange");
            textures[4] = game.Content.Load<Texture2D>("Textures//green");
            textures[5] = game.Content.Load<Texture2D>("Textures//lightblue");
            textures[6] = game.Content.Load<Texture2D>("Textures//prorange");
            textures[7] = game.Content.Load<Texture2D>("Textures//teal");
            displayTypes.Add(typeof(Box), typeof(DisplayBoxTest));
            displayTypes.Add(typeof(Sphere), typeof(DisplaySphereTest));
            displayTypes.Add(typeof(Capsule), typeof(DisplayCapsuleTest));
            displayTypes.Add(typeof(Cylinder), typeof(DisplayCylinderTest));
            displayTypes.Add(typeof(Cone), typeof(DisplayConeTest));
            displayTypes.Add(typeof(Triangle), typeof(DisplayTriangleTest));
            displayTypes.Add(typeof(ConvexHull), typeof(DisplayConvexHullTest));
            displayTypes.Add(typeof(MinkowskiSum), typeof(DisplayMinkowskiSumTest));
            displayTypes.Add(typeof(WrappedBody), typeof(DisplayWrappedBodyTest));
            displayTypes.Add(typeof(CompoundBody), typeof(DisplayCompoundBodyTest));
            displayTypes.Add(typeof(FluidVolume), typeof(DisplayFluidTest));
            displayTypes.Add(typeof(Terrain), typeof(DisplayTerrainTest));
            displayTypes.Add(typeof(TriangleMesh), typeof(DisplayTriangleMeshTest));

            fillState = new RasterizerState();
            wireframeState = new RasterizerState();
            wireframeState.FillMode = FillMode.WireFrame;
        }

        /// <summary>
        /// Constructs a new display object for an object.
        /// </summary>
        /// <param name="objectToDisplay">Object to create a display object for.</param>
        /// <returns>Display object for an object.</returns>
        internal ModelDisplayObjectBase getDisplayObject(object objectToDisplay)
        {
            Type displayType;
            if (!displayObjects.ContainsKey(objectToDisplay) &&
                myDisplayTypes.TryGetValue(objectToDisplay.GetType(), out displayType))
            {
#if !WINDOWS
                return (ModelDisplayObjectBase)displayType.GetConstructor(
                                                     new Type[] { typeof(ModelDrawer), objectToDisplay.GetType() })
                                                     .Invoke(new object[] { this, objectToDisplay });
#else
                return (ModelDisplayObjectBase)Activator.CreateInstance(displayType, new object[] { this, objectToDisplay });
#endif
            }
            return null;
        }

        /// <summary>
        /// Attempts to add an object to the ModelDrawer.
        /// </summary>
        /// <param name="objectToDisplay">Object to be added to the model drawer.</param>
        /// <returns>ModelDisplayObject created for the object.  Null if it couldn't be added.</returns>
        public ModelDisplayObjectBase add(object objectToDisplay)
        {
            ModelDisplayObjectBase displayObject = getDisplayObject(objectToDisplay);
            if (displayObject != null)
            {
                add(displayObject);
                displayObjects.Add(objectToDisplay, displayObject);
                return displayObject;
            }
            else
                return null; //Couldn't add it.
        }

        /// <summary>
        /// Adds the display object to the drawer.
        /// </summary>
        /// <param name="displayObject">Display object to add.</param>
        /// <returns>Whether or not the display object was added.</returns>
        public bool add(SelfDrawingModelDisplayObject displayObject)
        {
            if (!selfDrawingDisplayObjects.Contains(displayObject))
            {
                selfDrawingDisplayObjects.Add(displayObject);
                return true;
            }
            return false;
        }

        /// <summary>
        /// Adds a display object to the drawer.
        /// </summary>
        /// <param name="displayObject">Display object to add.</param>
        protected abstract void add(ModelDisplayObjectBase displayObject);

        /// <summary>
        /// Removes an object from the drawer.
        /// </summary>
        /// <param name="objectToRemove">Object to remove.</param>
        public void remove(object objectToRemove)
        {
            remove(displayObjects[objectToRemove]);
        }

        /// <summary>
        /// Removes an object from the drawer.
        /// </summary>
        /// <param name="displayObject">Display object to remove.</param>
        public void remove(SelfDrawingModelDisplayObject displayObject)
        {
            selfDrawingDisplayObjects.Remove(displayObject);
        }


        /// <summary>
        /// Removes a display object from the drawer.
        /// </summary>
        /// <param name="displayObject">Object to remove.</param>
        protected abstract void remove(ModelDisplayObjectBase displayObject);

        /// <summary>
        /// Cleans out the model drawer of any existing display objects.
        /// </summary>
        public void clear()
        {
            displayObjects.Clear();
            selfDrawingDisplayObjects.Clear();
            clearManagedModels();
        }

        /// <summary>
        /// Cleans out any data contained by derived drawers.
        /// </summary>
        protected abstract void clearManagedModels();

        /// <summary>
        /// Determines if the object has an associated display object in this drawer.
        /// </summary>
        /// <param name="displayedObject">Object to check for in the drawer.</param>
        /// <returns>Whether or not the object has an associated display object in this drawer.</returns>
        public bool contains(object displayedObject)
        {
            return displayObjects.ContainsKey(displayedObject);
        }

        /// <summary>
        /// Updates the drawer and its components.
        /// </summary>
        public void update()
        {
            foreach (SelfDrawingModelDisplayObject displayObject in selfDrawingDisplayObjects)
                displayObject.update();
            updateManagedModels();
        }

        /// <summary>
        /// Updates the drawer's technique.
        /// </summary>
        protected abstract void updateManagedModels();

        /// <summary>
        /// Draws the drawer's models.
        /// </summary>
        /// <param name="viewMatrix">View matrix to use to draw the objects.</param>
        /// <param name="projectionMatrix">Projection matrix to use to draw the objects.</param>
        public void draw(Matrix viewMatrix, Matrix projectionMatrix)
        {
            if (isWireframe)
                game.GraphicsDevice.RasterizerState = wireframeState;
            else
                game.GraphicsDevice.RasterizerState = fillState;

            game.GraphicsDevice.BlendState = BlendState.Opaque;
            game.GraphicsDevice.DepthStencilState = DepthStencilState.Default;

            foreach (SelfDrawingModelDisplayObject displayObject in selfDrawingDisplayObjects)
                displayObject.draw(viewMatrix, projectionMatrix);
            drawManagedModels(viewMatrix, projectionMatrix);

        }

        /// <summary>
        /// Draws the models managed by the drawer using the appropriate technique.
        /// </summary>
        /// <param name="viewMatrix">View matrix to use to draw the objects.</param>
        /// <param name="projectionMatrix">Projection matrix to use to draw the objects.</param>
        protected abstract void drawManagedModels(Matrix viewMatrix, Matrix projectionMatrix);
    }
}
